import numpy as np
import numpy.ma as mp
from scipy.interpolate import griddata
import logging

from xframe.library.mathLibrary import cartesian_to_spherical, spherical_to_cartesian, get_gaussian_weights_1d

log=logging.getLogger('root')

class Grid:
    types={'uniform':0,'uniform_dependend':1,'nonUniform_dependend':2,'custom':3}

    ###Class function###
    def copy(grid):
        if grid.gridType!=Grid.types['nonUniform_dependend']:
            copiedArray=np.copy(grid.array)
        else:
            copiedArray=mp.copy(grid.array)
        copiedGrid=Grid(copiedArray,gridType=grid.gridType)
        return copiedGrid

    
    def __init__(self,gridArray,gridType='custom'):    
        '''
        Assumes gridArray is a normal numpy array if dimension =1 or an array of dtype object which contains numpy arrays of type other than object which correspond to grid points.
        '''
         
        try:
            if type(gridType)==str:
                self.gridType=self.types[gridType]
            else:
                assert gridType in self.types.values()
                self.gridType=gridType
        except (KeyError,AssertionError):
            log.error('Given gridType ={} is not known!'.format(gridType))
            raise
        
        firstElement=gridArray.flatten()[0]
        if isIterable(firstElement):
            self.dimension=len(firstElement)
        else:
            self.dimension=0

        if self.gridType==self.types['nonUniform_dependend']:
            assert isinstance(gridArray,np.ma.core.MaskedArray),'Grid Type nonUniform expects MaskedArray gridArrays but type given was {}'.format(type(gridArray))
        self.array=gridArray
        self.shape=gridArray.shape
        
        self.getitem=self.generateGetItemFunction(self.gridType)
        self.setitem=self.generateSetItemFunction(self.gridType)
        self.applyFunctionToElements=self.generateApplyFunctionToElements(self.gridType)
        self.applyFunctionAlongAxis=self.generateApplyFunctionToAxis(self.gridType)

    def setDimension(self,newDimension):
        self.dimension=newDimension

        self.getitem=self.generateGetItemFunction(self.gridType)
        self.setitem=self.generateSetItemFunction(self.gridType)
        
    def setDimensionAndGridType(self,newGridType,newDimension):
        try:
            if type(gridType)==str:
                self.gridType=self.types[newGridType]
            else:
                assert newGridType in types.values()
                self.gridType=newGridType
        except (KeyError,AssertionError):
            log.error('Given gridType ={} is not known!'.format(newGridType))
            raise
        self.dimension=newDimension

        self.getitem=self.generateGetItemFunction(self.gridType)
        self.setitem=self.generateSetItemFunction(self.gridType)
        self.applyFunctionToElements=self.generateApplyFunctionToElements(self.gridType)
        self.applyFunctionAlongAxis=self.generateApplyFunctionToAxis(self.gridType)

    def setArray(self,newArray):
        self.array=newArray
#        log.info('new array.shape={}'.format(newArray))
        self.shape=newArray.shape
        
    def __getitem__(self,key):
        item=self.getitem(self,key)
        return item
    
    def __setitem__(self,key,value):
        self.setitem(self,key,value)

    def __mul__(self,multiplicant):
        if self.gridType==self.types['nonUniform_dependend']:
            self.array[~self.array.mask]*=multiplicant
        else:
            self.array*=multiplicant
        return self
            
    def apply_along_axis(self,function,axis,*args,**kwargs):
#        log.info('apply along axis ={}'.format(self.applyFunctionAlongAxis))
        sampledGrid=self.applyFunctionAlongAxis(self,function,axis,*args,**kwargs)
        return sampledGrid
    
    def apply(self,function,*args,**kwargs):
#        log.info('alive!')
        sampledGrid=self.applyFunctionToElements(self,function,*args,**kwargs)
        return sampledGrid

    def reshape(self,*args):
        self.array=self.array.reshape(*args)
        self.shape=self.array.shape
        return self

    def transpose(self,**kwargs):
        self.array=np.transpose(self.array,**kwargs)
        self.shape=self.array.shape
        return self
    
    def swapaxes(self,*args):
        self.array=self.array.swapaxes(*args)
        self.shape=self.array.shape
        return self
        
    def generateSetItemFunction(self,gridType):
        def setitem_nonUniform(self,key,value):
            array=self.array
            item=array.__getitem__(key)
            shape=item.shape
            mask=array.mask.__getitem__(key)
            keyIndexesSingleValue=isinstance(mask,np.bool_)

            if keyIndexesSingleValue:
                if not mask:
                    array.__setitem__(key,value)
                else:
                    log.info('trying to set a single masked value!')
            else:
                tempView=item.reshape(-1)
                validDataIndexes=tuple(np.arange(tempView.shape[0])[~tempView.mask])
                tempView[(validDataIndexes,)]=getArrayOfArray(value).flatten()
                array.__setitem__(key,item)

        def setitem_nonUniform1D(self,key,value):
            array=self.array
            item=array.__getitem__(key)
            shape=item.shape
            mask=array.mask.__getitem__(key)
            keyIndexesSingleValue=isinstance(mask,np.bool_)

            if keyIndexesSingleValue:
                if not mask:
                    array.__setitem__(key,value)
                else:
                    log.info('trying to set a single masked value!')
            else:
                tempView=item.reshape(-1)
                validDataIndexes=tuple(np.arange(tempView.shape[0])[~tempView.mask])
                tempView[(validDataIndexes,)]=value
                array.__setitem__(key,item)
                

        def setitem_uniform(self,key,value):
            array=self.array
            item=array.__getitem__(key)
#            log.info('item = {}'.format(item))
            keyIndexesSingleValue= (item.dtype!=np.object)
#            log.info('keyIndexesSingleValue={}'.format(keyIndexesSingleValue))
            if keyIndexesSingleValue:
                array.__setitem__(key,value)
            else:
                shape=item.shape
                tempView=item.flatten()
                tempView[:]=getArrayOfArray(value)
#                log.info('temp view={}'.format(tempView))
                array.__setitem__(key,tempView.reshape(shape))

        def setitem_uniform1D(self,key,value):
            array=self.array
            item=array.__getitem__(key)
            keyIndexesSingleValue= not isIterable(item)
            if keyIndexesSingleValue:                
                array.__setitem__(key,value)
            else:
                shape=item.shape
                tempView=item.reshape(-1)
                tempView[:]=value
                array.__setitem__(key,item)

                
        types=self.types
        dimension=self.dimension
        if gridType==types['nonUniform_dependend']:
            if dimension==0:
                setitem=setitem_nonUniform1D
            else:
                setitem=setitem_nonUniform
        else:
            if dimension==0:
                setitem=setitem_uniform1D
            else:
                setitem=setitem_uniform
        return setitem
            
    def generateGetItemFunction(self,gridType):
        def getitem_nonUniform(self,key):
            array=self.array
            keyIndexesSingleValue= isinstance(array.mask.__getitem__(key),np.bool_)
            if keyIndexesSingleValue:
                item=np.array(tuple(array.__getitem__(key)))
            else:
                item=array.__getitem__(key)
                item=np.array(tuple(item[~item.mask]))
            return item

        def getitem_nonUniform1D(self,key):
            array=self.array
            keyIndexesSingleValue= isinstance(array.mask.__getitem__(key),np.bool_)
            if keyIndexesSingleValue:
                item=array.__getitem__(key)
            else:
                item=array.__getitem__(key)
                item=np.array(item[~item.mask])
            return item
                
        def getitem_uniform(self,key):
            item=self.array.__getitem__(key)
            keyIndexesSingleValue= (item.dtype!=np.object)
#            log.info('key={}  singleValue={} item dtype={}'.format(key,keyIndexesSingleValue,item.dtype))
            if not keyIndexesSingleValue:
                item=item.flatten()
                item=np.array(tuple(item))
            return item
        
        def getitem_uniform1D(self,key):
            item=self.array.__getitem__(key)
            
            keyIndexesSingleValue= not isIterable(item)
  #          log.info('key={}  singleValue={}'.format(key,keyIndexesSingleValue))
            if not keyIndexesSingleValue:
                item=item.flatten()
            return item

        types=self.types
        dimension=self.dimension
        if gridType==types['nonUniform_dependend']:
            if dimension==0:
                getitem=getitem_nonUniform1D
            else:
                getitem=getitem_nonUniform
        else:
            if dimension==0:
                getitem=getitem_uniform1D
            else:
                getitem=getitem_uniform
        return getitem


    ###generate Apply Function along Axis###
    def generateApplyFunctionToAxis(self,gridType):   
        dimension=self.dimension
        if gridType==self.types['nonUniform_dependend']:
            if dimension==0:                
                decorator=concatenateDecorators([self.formatFunctionOutput,self.formatFunctionInput_applyMask])
                applyFunctionToAxis=self.plugDecoratorIntoTemplate(self.applyFunctionToAxis_nonUniform_template,decorator)
            else:
                decorator=concatenateDecorators([self.formatFunctionOutput,self.formatFunctionInput_nD,self.formatFunctionInput_applyMask])
                applyFunctionToAxis=self.plugDecoratorIntoTemplate(self.applyFunctionToAxis_nonUniform_template,decorator)
        else:
            if dimension==0:
                decorator=self.formatFunctionOutput
                applyFunctionToAxis=self.plugDecoratorIntoTemplate(self.applyFunctionToAxis_uniform_template,decorator)
            else:
                decorator=concatenateDecorators([self.formatFunctionOutput,self.formatFunctionInput_nD])
                applyFunctionToAxis=self.plugDecoratorIntoTemplate(self.applyFunctionToAxis_uniform_template,decorator)
        return applyFunctionToAxis

    def formatFunctionOutput(self,function):
        def newFunction(*args,**kwargs):
            output=function(*args,*kwargs)
            
            outputIsArray=isIterable(output)
            if outputIsArray:
                outputIsAnArrayOfArrys= len(output.shape)>1
                if outputIsAnArrayOfArrys:
                    output=getArrayOfArray(output)
            return output
        return newFunction

    def formatFunctionInput_nD(self,function):
        def newFunction(*args,**kwargs):
            #                log.info('input={}'.format(args[0]))
            args=(np.array(tuple(args[0])),)+args[1:]
            #                log.info('input2={}'.format(args[0]))
            output=function(*args,*kwargs)
            return output
        return newFunction

    def formatFunctionInput_applyMask(self,function):
        def newFunction(*args,**kwargs):
            args=(args[0][~args[0].mask],)+args[1:]
            output=function(*args,*kwargs)
            return output
        return newFunction
        
    def plugDecoratorIntoTemplate(self,template,decorator):            
        def applyFunctionAlongAxis(self,function,axis,*args,**kwargs):
#            log.info('function={}'.format(function))
#            log.info('template={}'.format(template))            
            output=template(decorator,function,axis,*args,**kwargs)
            return output
        return applyFunctionAlongAxis


    def applyFunctionToAxis_nonUniform_template(self,functionDecorator,function,axis,*args,**kwargs):
        function=appendArgumentsToFunction(function,*args,**kwargs)
        newFunction=functionDecorator(function)
        SampledFunction,newShape=getSampleFunctionAndShape(self,newFunction,axis)

        newArraySpecifier=generateNewArraySpecifier_nonUniform(SampledFunction,newShape)
        oldShape=self.array.shape
        newArray=getArray_nonUniform(newArraySpecifier)
        #        log.info('sampled Function={}'.format(SampledFunction))
        #        log.info('sampled Function2={}'.format(SampledFunction.reshape(-1,sampledGrid.dimension)))
        self.array=newArray
        newDimension=newArraySpecifier[1]
        newArrayIsUniform=newArraySpecifier[4]
        if newArrayIsUniform:
            newGridType=self.types['uniform_dependend']
        else:
            newGridType=self.types['nonUniform_dependend']
        self.setDimensionAndGridType(newGridType,newDimension)
        self[:]=SampledFunction.flatten()
        #       for index,gridPart in enumerate(self.array):
        #           gridPart[:len(SampledFunction[index])]=SampledFunction[index]
        self.swapaxes(axis,lastAxis)
        return self
    
    def applyFunctionToAxis_uniform_template(self,functionDecorator,function,axis,*args,**kwargs):
        function=appendArgumentsToFunction(function,*args,**kwargs)
        newFunction=functionDecorator(function)
        array=self.array
        newArray=np.apply_along_axis(newFunction,axis,array)
#        log.info('new shape={}'.format(newArray.shape))
        dimension=np.abs(len(newArray.shape)-len(array.shape))
        self.array=newArray
        self.setDimension(dimension)
        self.shape=newArray.shape
#        log.info('dimension={}'.format(self.dimension))
        return self
            

    def getSampleFunctionAndShape(self,function,axis):            
        array=self.array
        lastAxis=len(originalShape)-1
        
#        log.info('last axis={}'.format(lastAxis))
        if lastAxis!=0:
            if axis!=lastAxis:
                array=array.swapaxes(axis,lastAxis)
            shape=array.shape
            array=array.reshape(-1,originalShape[axis])
        else:
            shape=array.shape
            array=mp.expand_dims(array,0)
            
        SampledFunction=np.array(list(map(function,array)))                
#        log.info('sampled Function.shape={}'.format(SampledFunction.shape))
        return SampledFunction,shape

    def generateNewArraySpecifier_nonUniform(self,SampledFunction,shape):
        def generateMask(axesLength,newShape):
            mask=np.full(newShape,True)
            for index,axis in enumerate(mask):
                axis[:axesLength[index]]=False
            return mask
        
        shapeWithoutAxis=shape[:-1]            
        nrOfAxes=np.sum(shapeWithoutAxis)
        mask=False
        if isIterable(SampledFunction[0]):
            axesLength=np.array(tuple(map(len,SampledFunction)))
            maxAxisLength=np.max(axesLength)
            minAxisLength=np.min(axesLength)

            uniform=False
            if maxAxisLength==minAxisLength:
                uniform=True
                
            if maxAxisLength>1:
                newShape=shapeWithoutAxis+(maxAxisLength,)
                if not uniform:
                    mask=generateMask(axesLength,newShape)
                    
            elif maxAxisLength==1:
                newShape=shapeWithoutAxis+(1,)
                uniform=True
            elif maxAxisLength==0:
                newShape=shapeWithoutAxis
                    
            if isIterable(SampledFunction[0][0]):
                dimension=len(SampledFunction[0][0])
                newDtype=np.object
            else:
                dimension=0
                newDtype=SampledFunction[0].dtype
        else:
            dimension=0
            newShape=shapeWithoutAxis
            newDtype=SampledFunction.dtype
        log.info('new Datatype={}, new Dimension={}'.format(newDtype,dimension))
        return [newShape,dimension,newDtype,mask,uniform]

    def getArray_nonUniform(self,newArraySpecifier):                
        newShape,dimension,newDtype,mask,uniform=newArraySpecifier

        if dimension==0:
            newArray=np.full(newShape,-9999,dtype=newDtype)
        else:
            newArray=np.full(newShape+(dimension,),-9999,dtype=newDtype)
            newArray=getArrayOfArray(newArray)
                
        if not uniform:
            newArray=mp.array(newArray,mask=mask)
        return newArray
        
        
    
    ### generate Apply Function To Elements###
    def generateApplyFunctionToElements(self,gridType):        
        applyFunctionToElements=self.plugInArrayCreationRoutine(self.applyFunctionToElements_template,gridType)
        return applyFunctionToElements
    
    def plugInArrayCreationRoutine(self,applyFunctionToElements_template,gridType):
        if gridType==self.types['nonUniform_dependend']:
            creationRoutine=self.arrayCreation_nonUniform
        else:
            creationRoutine=self.arrayCreation_uniform
        def applyFunctionToElements(self,function,*args,**kwargs):
            sampledGrid=self.applyFunctionToElements_template(creationRoutine,function,*args,**kwargs)
            return sampledGrid
        return applyFunctionToElements    

    def arrayCreation_uniform(self,dimension,newDtype):
        shape=self.array.shape
        if dimension==0:
            newArray=np.full(shape,-9999,dtype=newDtype)
        else:
            newArray=getArrayOfArray(np.full(shape+(dimension,),-9999,dtype=newDtype))
            #            log.info('newArray={}'.format(newArray))
        return newArray
        
    def arrayCreation_nonUniform(self,dimension,newDtype):
        oldArray=self.array
        shape=oldaArray.shape
        mask=oldArray.mask
        if dimension==0:
            newArray=mp.array(np.full(shape,-9999,dtype=newDtype),mask=mask)
        else:
            unMaskedArray=getArrayOfArray(np.full(shape+(dimension,),-9999,dtype=newDtype))
            newArray=mp.array(unMaskedArray,mask=mask)
        return newArray

    def applyFunctionToElements_template(self,arrayCreationRoutine,function,*args,**kwargs):
        function=appendArgumentsToFunction(function,*args,**kwargs)
        SampledFunction=np.array(list(map(function,self[:])))
#        log.info('self shape={}'.format(self[:].shape))
#        log.info('SampledFunction shape={}'.format(SampledFunction.shape))
        #          log.info('function= {}'.format(SampledFunction))
        dimension,newDtype=self.generateDimensionAndDtypeOfNewArray_elements(SampledFunction)

        newArray=arrayCreationRoutine(dimension,newDtype)
        self.array=newArray
        self.setDimension(dimension)
        self.shape=newArray.shape
        self[:]=SampledFunction
        return self
    def generateDimensionAndDtypeOfNewArray_elements(self,SampledFunction):
        try:
            dimension=len(SampledFunction[0])                
            newDtype=SampledFunction[0].dtype
        except TypeError:
            dimension=0
            newDtype=SampledFunction.dtype
            #            log.info('new Datatype={}, new Dimension={}'.format(newDtype,dimension))
        return dimension,newDtype

class NestedArray:
    def __init__(self,array,n_ndim):        
        if isinstance(array,np.ndarray):
            self.is_masked=False
        elif isinstance(array,np.ma.core.MaskedArray):
            self.is_masked=True
        else:
            e=AssertionError('Array has to be a numpy array of type "ndarray" ore "MaskedArray". Provided array type="{}"'.format(type(array)))
            log.error(e)
            raise e

        try:
            assert array.ndim>=n_ndim, '(ndim of subarray) ={} has to be less or equal to (ndim of complete array) ={}'.format(n_ndim,array.ndim)
        except AssertionError as e:
            log.error(e)
            raise

        self._array=array
        self._n_ndim=n_ndim

        
    def __getitem__(self,key):
        return self._array.__getitem__(key)
    
    def __setitem__(self,key,value):
        self._array.__setitem__(key,value)

    def __mul__(self,value):
        return self._array.__mul__(value)                
    def __add__(self,value):
        return self._array.__add__(value)
    def __pow__(self,value):
        return self._array.__pow__(value)
                
    @property
    def array(self):
        return self._array
    @property
    def array1d(self):
        return self._array.reshape(-1,*self.n_shape)
    @property
    def total_ndim(self):
        return self._array.ndim
    @property
    def ndim(self):
        return self._array.ndim-self._n_ndim
    @property
    def n_ndim(self):
        return self._n_ndim
    @property
    def total_shape(self):
        return self._array.shape
    @property
    def shape(self):
        return self.total_shape[:self.ndim]
    @property
    def n_shape(self):
        return self._array.shape[self.ndim:]
    

    @array.setter
    def array(self,new_array):
        try:
            assert new_array.ndim>=self._n_ndim, '(ndim of new array)={} has to be greater or equal to (ndim of subarrays) ={}'.format(len(new_array.shape),self._n_ndim)
        except AssertionError as e:
            log.error(e)
            raise e
        self._array=new_array
    @n_ndim.setter
    def n_ndim(self,new_nested_ndim):
        try:
            assert self.total_ndim>=new_nested_ndim, '(new ndim of subarrays) ={} has to be less or equal to (ndim of total array) ={}'.format(new_nested_ndim,self.total_ndim)
        except AssertionError as e:
            log.error(e)
            raise e
        self._n_ndim=new_nested_ndim
        self.apply=self.generate_apply_method()

    def copy(self):
        array_copy=self._array.copy()
        return NestedArray(array_copy,self._n_ndim)
    def flatten(self):
        return NestedArray(self._array.reshape(-1,self.n_shape[0]),self._n_ndim)
    
    def apply(function):
        array=self._array
        n_ndim=self._n_ndim
        ndim=array.ndim-n_ndim
        value=func(array,n_ndim)
        new_n_ndim=value.ndim-self.ndim

        if value.shape[:ndim]==array.shape[:ndim]:
            e=AssertionError('Applying Function {} changed non-nested shape.'.format(func))
            log.error(e)
            raise e
        
        self._array=value
        self._n_ndim=new_n_ndim  

class CoordinateSystems:
    coord_systems=('spherical','cartesian')
    conversion_routines={coord_systems[0]:{coord_systems[1]:spherical_to_cartesian},coord_systems[1]:{coord_systems[0]:cartesian_to_spherical}}
class ReGrider(CoordinateSystems):
    @classmethod
    def regrid(cls,data,grid,coordinate_type,new_grid,new_coordinate_type,options={}):
        systems=cls.coord_systems        
        try:
            assert (new_coordinate_type in systems) and (coordinate_type in systems),'current and new coordinate types "{}" are not known. Known types are {}. '.format((new_coordinate_type,coordinate_type),systems)
            if (new_coordinate_type in systems) ^ (coordinate_type in systems):
                if (coordinate_type in systems):
                    log.error( 'new coordinate type "{}" is not known. Known types are {}. Continue with current = new type '.format(new_coordinate_type,systems))
                    new_coordinate_type=coordinate_type
                else:
                    log.error( 'coordinate type "{}" is not known. Known types are {}. Continue with current = new type '.format(coordinate_type,systems))
                    coordinate_type=new_coordinate_type
        except AssertionError as e:
            log.error(e)
            raise

            
        if coordinate_type == new_coordinate_type:
            if coordinate_type==cls.coord_systems[0]:
                method=cls.regrid_spher
            else:
                method=cls.regrid_cart
        elif coordinate_type==systems[1] and new_coordinate_type==systems[0]:
            method=cls.regrid_cart_to_spher
        elif coordinate_type==systems[0] and new_coordinate_type==systems[1]:
            method=cls.regrid_spher_to_cart
        #log.error('initial data shape ={}'.format(data.shape))
        #log.error('initial grid shape ={}'.format(grid.total_shape))
        #log.error('new grid shape ={}'.format(new_grid.total_shape))


        if len(data.shape) == 1:
            apply_over_axis = False
        else:
            apply_over_axis = options.get('apply_over_axis',False)
            
        #log.info('apply over axis = {}'.format(apply_over_axis))
        if isinstance(apply_over_axis,bool):         
            new_data=method(data,grid,new_grid,options)
        else:
            application_axis=apply_over_axis            
            reordered_data=np.moveaxis(data,application_axis,0)
            #log.error('reordered_data shape ={}'.format(reordered_data.shape))            
            new_data=np.array(tuple(method(tmp_data,grid,new_grid,options) for tmp_data in reordered_data))
            new_data=np.moveaxis(new_data,0,application_axis)
        return new_data

    @classmethod
    def regrid_cart(cls,data,grid,new_grid,options):
        systems=cls.coord_systems
        dims=new_grid.n_shape[0]
        #log.info('dims={}'.format(dims))
        fill_method = options.get('fill_method', 'constant')
        griddata_opt={'rescale':options.get('rescale',False),'method':options.get('interpolation','nearest'),'fill_value':options.get('fill_value',np.nan)}
        #option to fill missing data by fitting gaussian to existing data
        if fill_method =='gaussian_fit':
            gaussian=cls.generate_gaussian(cls.coord_sys)
            griddata_opt['fill_value']=np.nan
        shape=new_grid.shape
        #        log.error('dims={}'.format(dims))
        if dims == 1:
            #log.info('old_grid shape ={} new grid shape ={} data shape = {}'.format(grid.total_shape,new_grid.total_shape,data.shape))
            new_data=griddata(grid[:],data,new_grid[:], **griddata_opt)
        else:
            new_data=griddata(grid.array.reshape(-1,dims),data.reshape(-1),new_grid.array.reshape(-1,dims), **griddata_opt)
        new_data=new_data.reshape(shape)        
        if fill_method =='gaussian_fit':
            filled_data_mask=(new_data==np.nan)            
            new_data[filled_data_mask]=gaussian(new_grid[filled_data_mask])
        return new_data

    @classmethod
    def regrid_spher(cls,data,grid,new_grid,options):
        systems=cls.coord_systems
        dims=new_grid.n_shape[0]
        fill_method_dict=options.get('fill_method',{'id':'constant'})
        griddata_opt={'rescale':options.get('rescale',False),'method':options.get('interpolation','linear'),'fill_value':fill_method_dict.get('value',np.nan)}
        if fill_method_dict['id']=='gaussian_fit':
            gaussian=cls.generate_gaussian(cls.coord_sys)
            griddata_opt['fill_value']=np.nan
            
        to_cart=cls.conversion_routines[systems[0]][systems[1]]
        to_spher=cls.conversion_routines[systems[1]][systems[0]]
        grid_c = to_cart(grid)
        new_grid_c = to_cart(new_grid)
        new_data=griddata(grid_c.array.reshape(-1,dims),data.reshape(-1), new_grid_c.array.reshape(-1,dims), **griddata_opt)
        new_data=new_data.reshape(new_grid.shape)
        
        if fill_method_dict['id']=='gaussian_fit':
            filled_data_mask=(new_data==np.nan)            
            new_data[filled_data_mask]=gaussian(new_grid[filled_data_mask])
#        cls.data=new_data
        return new_data

    @classmethod
    def regrid_cart_to_spher(cls,data,grid,new_grid,options):
        log.error('grid={}'.format)
        grid=grid.copy()
        dims=grid.n_shape[0]
        fill_method_dict=options.get('fill_method',{'id':'constant'})
        griddata_opt={'rescale':options.get('rescale',False),'method':options.get('interpolation','linear'),'fill_value':fill_method_dict.get('value',np.nan)}
        
        max_dim=np.max(grid[:])
        cartesian_to_spherical(grid)
        trf_support_mask=(grid[...,0]<=max_dim)
        grid=grid[trf_support_mask]
        data=data.copy()[trf_support_mask]
        spherical_to_cartesian(grid)
        new_grid_c=new_grid.copy()
        spherical_to_cartesian(new_grid_c)
        new_data=griddata(grid.reshape(-1,dims),data.reshape(-1),new_grid_c.array.reshape(-1,dims),**griddata_opt)
        new_data=new_data.reshape(new_grid.shape)
        return new_data

    @classmethod
    def regrid_spher_to_cart(cls,data,grid,new_grid,options):
        systems=cls.coord_systems
        grid=grid.copy()
        dims=grid.n_shape[0]
        fill_method_dict=options.get('fill_method',{'id':'constant'})
        griddata_opt={'rescale':options.get('rescale',False),'method':options.get('interpolation','linear'),'fill_value':fill_method_dict.get('value',np.nan)}        
        max_r=np.max(grid[...,0])       
        gaussian=cls.generate_gaussian(data,grid,systems[0],systems[1],options)
        
        spherical_to_cartesian(grid)
        new_data=griddata(grid.array.reshape(-1,dims),data.reshape(-1),new_grid.array.reshape(-1,dims),**griddata_opt)
        new_data=new_data.reshape(new_grid.shape)
        trf_support_mask=np.prod(tuple(new_grid[...,dim]<=max_r for dim in np.arange(dims)),axis=0).astype(bool)*(new_data==np.nan)
        new_data[trf_support_mask]=gaussian(new_data[trf_support_mask])
        return new_data

    @classmethod
    def generate_gaussian(cls,data,grid,coord_system,new_coord_sys,options={}):
        systems=cls.coord_systems
        bins=options.get('radial_bins',int(np.prod(data[0,...].shape)/10))
        sigma=cls.get_gaussian_sigma(data,grid,coord_system,bins)
        amplitude=np.max(data)
        if new_coord_sys==systems[0]:
            def gaussian(points):
                radi=points[...,0]
                return amplitude*np.exp(-(radi/sigma)**2)
        else:
            def gaussian(points):
                radi=np.linalg.norm(points[:],axis=-1)
                return amplitude*np.exp(-(radi/sigma)**2)
        return gaussian

    @classmethod
    def get_gaussian_sigma(cls,data,grid,current_coord_system,bins):
        systems=cls.coord_systems
        sys=current_coord_system

        if sys==systems[1]:            
            grid_radi=np.linalg.norm(grid[:],axis=-1)
        else:
            grid_radi=grid[...,0]
            
        pos_so_averaged_data,radi_edges=np.histogram(data,bins=bins,weights=grid_radi)
        pos_radi=radi_edges[:-1]+np.diff(radi_edges)/2
        radi=np.concatenate((-pos_radi[::-1],pos_radi))
        so_averaged_data=np.concatenate((pos_so_averaged_data[::-1],pos_so_averaged_data))                
        sigma=get_gaussian_weights_1d(so_averaged_data,radi)['sigma']
        return sigma
    

class SampledFunction(CoordinateSystems):
    def __init__(self,grid,data,coord_sys='spherical'):
        self.initial_grid=grid.copy()
        try:
            assert coord_sys in self.coord_systems, 'Unknown coord_sys "{}". Has to be "cartesian" or "spherical".'.format(coord_sys)
        except AssertionError as e:
            log.error(e)
            raise e
        self.initial_coord_sys=coord_sys
        self.initial_data=data.copy()
        self.grid=grid.copy()
        self.coord_sys=self.initial_coord_sys
        self.data=data.copy()
        
    def __getitem__(self,key):
        return self.data.__getitem__(key)
    
    def __setitem__(self,key,value):
        self.data.__setitem__(key,value)

    def override_grid(self,grid):
        self.initial_grid=grid.copy()
        self.grid=grid
        
        
    #uses ReGrider class
    def regrid(self,new_grid,options={}):
        systems=self.coord_systems
        coordinate_type=self.initial_coord_sys        
        new_coordinate_type=options.get('new_coordinate_type',coordinate_type)
        try:
            assert new_coordinate_type in systems,'new coordinate type "{}" is not known. Known types are {}. Assume new = old coordinate type and continue.'.format(new_coordinate_type,systems)
        except AssertionError as e:
            log.error(e)
            new_coordinate_type=coordintae_type

        #log.info('aint before = {}'.format(self.data[:10]))
        #log.info('aint grid before = {}'.format(self.initial_grid[:10]))
        self.data=ReGrider.regrid(self.initial_data,self.initial_grid,coordinate_type,new_grid,new_coordinate_type,options=options)
        #log.info('aint grid after = {}'.format(new_grid[:10]))
        #log.info('aint after = {}'.format(self.data[:10]))
        self.grid=new_grid
        self.coord_sys=new_coordinate_type        

            

    def apply(self,function):
        self.data.apply(function)

        
class gridFactory:
    gridTypes=Grid.types
    
    def constructGrid(self,gridFunctionList,gridType):
        gridArray=self.constructGridArrayAndShape(gridFunctionList,gridType)
        grid=Grid(gridArray,gridType=gridType)
        return grid

    def constructGridArrayAndShape(self,gridFunctionList,gridType):
        types=self.gridTypes
        try:
            Type=types[gridType]
        except KeyError:
            log.error('Given gridType: {} is not known!'.format(gridType))
            raise
            
        if Type==types['uniform']:
            gridArray=self.constructGridArrayAndShape_Uniform(gridFunctionList)
        elif gridType==types['custom']:
            gridArray,gridShape=self.constructGridArrayAndShape_custom(gridFunctionList)
        else:
            gridArray,gridShape=self.constructGridArrayAndShape_Dependend(gridFunctionList,gridType)

        return gridArray

    def constuctGridArrayAndShape_custom(self,gridFunctionList):
        #not implemented yet
        return np.array([]),np.array([])
    
    def constructGridArrayAndShape_Uniform(self,gridFunctionList):
        def generateShape(dimensions):
            def repeatFunction(subList):
                numberOfRepeats=np.prod(subList[:-1])
                return np.repeat(subList[-1],numberOfRepeats)
            dimLengths=self.getDimensionSizeNumbers(dimensions)
            subListsOfDimLengths=list(map(lambda x : dimLengths[:x],np.arange(1,len(dimLengths)+1,1)))
            shape=list(map(repeatFunction,subListsOfDimLengths))
            return shape

        if isinstance(gridFunctionList[0],np.ndarray):
            dimensions=gridFunctionList
        else:
            dimensions=list(map(lambda func: func(),gridFunctionList))
        gridArray=generateMesh(*dimensions,joiningDimension=len(dimensions))
        if len(dimensions)!=1:
            gridArray=getArrayOfArray(gridArray)
        return gridArray

    def constructGridArrayAndShape_Dependend(self,gridFunctionList,gridType):
        
        def generateDimension(gridFunction,previousGridPart=[]):

            def generateDimension_Dependend(previousGridPart,gridFunction):
                elementNumbers=np.arange(len(previousGridPart))
                dimension=np.array(list(map(gridFunction,previousGridPart,elementNumbers)))
                return dimension

            def generateInitialDimension(initialGridFunction):
                dimension=initialGridFunction()
                dimension=np.expand_dims(dimension,axis=1)
                return dimension

            
            if previousGridPart==[]:
                dimension=generateInitialDimension(gridFunction)
            else:
                dimension=generateDimension_Dependend(previousGridPart,gridFunction)
            return dimension

        def appendDimensionListToGridPart_Uniform(part,dimList):
            newDimLength=len(dimList[0])
            repeatedGridPart=np.repeat(part,newDimLength,axis=0)
            newGridPart=np.concatenate((repeatedGridPart,np.expand_dims(dimList.flatten(),axis=1)),axis=1)
            return newGridPart

        def appendDimensionListToGridPart_NonUniform(part,dimList):
            def appendDimensionToGridPart(part,partLength,dim):
                dimLength=len(dim)
                repeatedGridPart=np.full((dimLength,partLength),part)
                newGridPart=np.concatenate((repeatedGridPart,np.expand_dims(dim,axis=1)),axis=1)
                newGridPart=np.split(newGridPart.flatten(),dimLength)
                return newGridPart
            
            subPartLength=len(part[0])
            partLength=len(part)
            subPartLengthIterator=repeat(subPartLength,partLength)
            newGridPart=list(map(appendDimensionToGridPart,part,subPartLengthIterator,dimList))
            newGridPart=flattenList(newGridPart)
            return newGridPart

        
        if gridType==self.gridTypes['uniform_dependend']:
            appendDimensionListToGridPart=appendDimensionListToGridPart_Uniform
        else:
            appendDimensionListToGridPart=appendDimensionListToGridPart_NonUniform

        gridShape=[]
        initialDimension=generateDimension(gridFunctionList[0])
        previousGridPart=initialDimension
        gridShape.append(np.array([ len(initialDimension)] ))
        for func in gridFunctionList[1:]:
            newDimension=generateDimension(func,previousGridPart=previousGridPart)
            gridShape.append(self.getDimensionSizeNumbers(newDimension))
            newGridPart=appendDimensionListToGridPart(previousGridPart,newDimension)
            previousGridPart=newGridPart
        gridArray=np.array(previousGridPart)
        return gridArray,gridShape

    def getDimensionSizeNumbers(self,dimension):
        sizes=np.array(list(map(len,dimension)))
        return sizes
    
class GridFactory:
    grid_types={'uniform':0,'uniform_dependent':1,'dependent':2}
    
    @classmethod
    def construct_grid(cls,grid_type,dimensions_list):
        grid_array=cls.constructGridArrayAndShape(grid_type,dimensions_list)
        grid=NestedArray(grid_array,1)
        return grid
    
    @classmethod
    def constructGridArrayAndShape(cls,grid_type,dimensions_list):
        types=cls.grid_types
        try:
            selected_type=types[grid_type]
        except KeyError:
            log.error('Given gridType: {} is not known!'.format(grid_type))
            raise
            
        if selected_type==types['uniform']:
            grid_array=cls.constructGridArray_Uniform(dimensions_list)
        else:
            grid_array=cls.constructGridArrayAndShape_Dependent(selected_type,dimensions_list)

        return grid_array
    
    @classmethod
    def constructGridArray_Uniform(cls,dimensions_list):
        first_dimension=dimensions_list[0]
        if callable(first_dimension):
            dimensions_list=tuple(dimension_func() for dimension_func in dimensions_list)
        n_lists = len(dimensions_list)
        ndim = sum([dim.ndim for dim in dimensions_list])
        
        if n_lists == ndim:
            grid_array=np.stack(np.meshgrid(*dimensions_list,indexing='ij'),ndim)
        else:
            shapes = tuple(dim.shape for dim in dimensions_list)
            shape = sum(shapes,())
            grid_array = np.zeros(shape,dtype = float)
            for n in range(n_lists):
                grid_array[...,]
            
        return grid_array

    @classmethod
    def constructGridArrayAndShape_Dependent(cls,grid_type,dimensions_list):
        
        def generate_dimension(dimension_func,grid_part=False):            
            is_first_part = isinstance(grid_part,bool)
            if callable(dimension_func):
                if is_first_part:
                    dimension=dimension_func()[:,None]
                else:
                    dimension=dimension_func(grid_part)
                if dimension.ndim == 1:
                    dimension=dimension[...,None]
            else:
                dimension = dimension_func
                if dimension.ndim == 1:
                    dimension=dimension[...,None]
                if not is_first_part:
                    array = np.zeros(grid_part.shape[:-1]+dimension.shape,dtype = dimension.dtype)
                    array[:]=dimension
                    dimension = array
            return dimension

        def append_dimension_uniform_dependent(grid_part,new_dimension):
            #dim_length=new_dimension.shape[-1]
            dim_length=new_dimension.shape[-2]
            grid_part_shape=grid_part.shape
            repeated_grid_part=np.repeat(grid_part,dim_length,-2).reshape(grid_part_shape[:-1]+(dim_length,)+grid_part_shape[-1:])
            #print('repeat shape = {}'.format(repeated_grid_part.dtype))
            #print('append shape = {}'.format(new_dimension.dtype))
            new_grid_part=np.concatenate((repeated_grid_part,new_dimension),axis=-1)
            return new_grid_part
        def append_dimension_dependent(grid_part,new_dimension):
            dim_length=new_dimension.shape[-1]
            grid_part_shape=grid_part.shape
            repeated_grid_part=mp.repeat(grid_part,dim_length,-2).reshape(grid_part_shape[:-1]+(dim_length,)+grid_part_shape[-1:])
            new_grid_part=mp.concatenate((repeated_grid_part,mp.expand_dims(new_dimension,-1)),axis=-1)
            
            vect_size=new_grid_part.shape[-1]
            new_mask=np.kron(np.sum(new_grid_part.mask,axis=-1,dtype=bool),np.full(vect_size,True))
            new_grid_part.mask=new_mask
            return new_grid_part

        
        dimensions_length=[]
        grid_shape=[]
        mask_part=np.array([1]*len(dimensions_list),dtype=int)
        first_dimension=generate_dimension(dimensions_list[0])
        grid_part=first_dimension
        grid_shape.append( (len(first_dimension),) )
        for dimension_func in dimensions_list[1:]:
            new_dimension=generate_dimension(dimension_func,grid_part=grid_part)
            if grid_type==cls.grid_types['dependent']:
                new_grid_part=append_dimension_dependent(grid_part,new_dimension)
            else:
                new_grid_part=append_dimension_uniform_dependent(grid_part,new_dimension)

            grid_part=new_grid_part
        gridArray=grid_part
        return gridArray


        
def double_dimension(grid,dimension):
    #doubles dimension of a grid and adds it in the position after the given dimension. 
    def generate_array(grid,npmp):
#        log.info('application dimension={}'.format(dimension))
        array=grid.array
        n_points=grid.shape[dimension]
        n_dimensions=grid.total_shape[-1]
#        log.info('n_dimensions={}'.format(n_dimensions))
        shape=grid.total_shape
        dimension_indexes= (...,)+(0,)*len(shape[dimension+1:-1])+(dimension,)
#        log.info('dim_indexes={}'.format(dimension_indexes))
        dim_values=array.__getitem__(dimension_indexes)
#        log.info('dim_values_2={}'.format(dim_values))

        repeat_axis=dimension        
        repeat_times=npmp.prod(shape[dimension:-1])
#        log.info('repeat times={}'.format(repeat_times))
        repeated_dim_values=npmp.repeat(dim_values[None,...,None],repeat_times,axis=dimension)
        repeated_dim_values=npmp.moveaxis(repeated_dim_values,-2,0).reshape(n_points,-1,1)
#        log.info('repeated dim values={}'.format(repeated_dim_values))        
        repeated_array=npmp.repeat(array[None,:],n_points,axis=0).reshape(n_points,-1,n_dimensions)
#        log.info('repeated_array={}'.format(repeated_array))
        
        wrong_axis_array=npmp.concatenate((repeated_dim_values,repeated_array),axis=2)
#        log.info('new_array ={}'.format(wrong_axis_array))
        dimension_mask=tuple(range(1,dimension+2))+(0,)+tuple(range(dimension+2,n_dimensions+1))
#        log.info('dimension mask ={}'.format(dimension_mask))
        wrong_shape_array=wrong_axis_array[...,dimension_mask]
#        log.info('wrong shape array={}'.format(wrong_shape_array))
        new_shape=(shape[dimension],)+shape[:dimension+1]+shape[dimension+1:-1]+(shape[-1]+1,)
#        log.info('new shape ={}'.format(new_shape))
        new_array=npmp.moveaxis(wrong_shape_array.reshape(*new_shape),0,dimension+1)
#        log.info('new array={}'.format(new_array))
        return new_array
    
    is_masked=isinstance(grid.array,np.ma.MaskedArray)
    if not is_masked:
        new_array=generate_array(grid,np)
    else:     
        new_array=generate_array(grid,np.ma)
        new_mask=(new_array,grid.ndim+1)
        new_array.mask=new_mask
    new_grid=NestedArray(new_array,1)
    return new_grid

def double_first_dimension(grid):
    def generate_array(grid,npmp):
        total_shape=grid.total_shape
        contracted_shape=(total_shape[0],np.prod(total_shape[1:-1]),total_shape[-1])
        first_dimension=grid.array.reshape(contracted_shape)[:,0,0]
        dim_length=len(first_dimension)
        
        new_dimension=npmp.kron(np.expand_dims(np.full(grid.shape,1),-1),first_dimension)
        
        repeated_grid=npmp.repeat(grid.array,dim_length,-2).reshape(total_shape[:-1]+(dim_length,)+total_shape[-1:])
#        log.info('repeated grid shape={}'.format(repeated_grid.shape))
        new_grid_array=npmp.moveaxis(npmp.concatenate((np.expand_dims(new_dimension,-1),repeated_grid),axis=-1),-2,0)
        return new_grid_array
    def generate_mask(new_array,new_grid_dim):
        new_mask=np.kron(np.prod(new_array.mask,-1,dtype=bool),np.full(new_grid_dim,True)).astype(bool)
        return new_mask

    is_masked=isinstance(grid.array,np.ma.MaskedArray)
    if not is_masked:
        new_array=generate_array(grid,np)
    else:     
        new_array=generate_array(grid,np.ma)
        new_mask=(new_array,grid.ndim+1)
        new_array.mask=new_mask
        
    new_grid=NestedArray(new_array,1)
    return new_grid

def uniformGrid_func(domain,nPoints='not specified',stepSize='not specified',noEndpoint=False):
    def uniformGridViaNPoints(domain,nPoints):
        def gridFunc(*args):
            if nPoints<=1:
                stepSize=0
            else:
                if noEndpoint:
                    stepSize=(domain[1]-domain[0])/nPoints
                else:
                    stepSize=(domain[1]-domain[0])/(nPoints-1)
#                    log.info('stepSize={}'.format(stepSize))
            gridArray=np.arange(nPoints,dtype=float)*stepSize+domain[0]
#            log.info('gridArray={}'.format(gridArray))
            return gridArray
        return gridFunc
    
    def uniformGridViaStepSize(domain,stepSize):
        def gridFunc(*args):
            if domain[0]==domain[1]:
                nPoints=1
            else:
                if noEndpoint:
                    nPoints=np.floor((domain[1]-domain[0])/stepSize)
                else:                        
                    nPoints=np.floor((domain[1]-domain[0])/stepSize)+1
            gridArray=np.arange(nPoints,dtype=float)*stepSize+domain[0]
            return gridArray
        return gridFunc
    def checkArguments(nPoints,stepSize):
        try:
            assert (nPoints!='not specified' or stepSize!='not specified')
        except AssertionError:
            log.warning( 'uniformGrid_func needs either a stepSize or nPoints as input.')

    checkArguments(nPoints,stepSize)
    gridFunc=[]
    if (nPoints!='not specified'):
        if stepSize=='not specified':
            gridFunc=uniformGridViaNPoints(domain,nPoints)
        else:
            log.warning('Number of steps as well as step size is specified. Defaulting to claculate grid for the number of Steps.')
            gridFunc=uniformGridViaNPoints(domain,nPoints)
    else:
        gridFunc=uniformGridViaStepSize(domain,stepSize)

    return gridFunc
    
def uniformGridGetStepSizes(gridArray):
    gridShape=gridArray.shape
#    log.info('grid shape={}'.format(gridShape))
    if gridArray.dtype==np.object:
        gridDimension=len(gridShape)
    else:
        gridDimension=gridShape[-1]

#    log.info('grid Dimension={}'.format(gridDimension))

    secondPointIndices=np.eye(gridDimension,dtype=np.int)
    thirdPointIndices=secondPointIndices*2
    stepSizes=gridArray[tuple(thirdPointIndices)]-gridArray[tuple(secondPointIndices)]
    indices=np.repeat(np.expand_dims(np.arange(gridDimension),0),2,axis=0)
#    log.info('indices={}'.format(indices))
#    log.info('steps={}'.format(np.array(tuple(stepSizes))))
    stepSizes=np.array(tuple(stepSizes))[tuple(indices)]
#    log.info('stepsizes dtype={}'.format(stepSizes.dtype))
    return stepSizes

def getGridByXYValues(xValues,yValues):
    def xfunc(*args):
        return xValues
    def yfunc(*args):
        return yValues
    gfact=gridFactory()
    newGrid=gfact.constructGrid([xfunc,yfunc],'uniform')
    return newGrid


def get_linspace_log2(domain, target_step_size,additional_steps = 0, return_parameters = False):
    try:
        assert (domain[0]<=0) and (domain[1]>0), 'domain: ( {} ) does not contain 0.'.format(domain)
    except AssertionError as e:
        log.error(e)
        traceback.print_exc()
        raise e
    domain = np.array(domain)
    length = domain[1]-domain[0]
    target_number_of_steps = length/target_step_size
    n_steps = 2**int(np.round(np.log2(target_number_of_steps)))+additional_steps
    
    partition = np.abs(np.ceil(domain/length*n_steps+0.5)).astype(int) #round up
    if partition[0] == partition[1]:
        partition[0] -= 1
    if domain[0] !=0:
        step_sizes = np.abs( (domain[0]/(partition[0]-1),domain[1]/(partition[1]-1)) )
        exact_domain_id = np.argmin(step_sizes) 
        step_size = step_sizes[exact_domain_id]
        log.info('step_size = {}'.format(step_size))
    else:
        step_size = domain[1]/(partition[1]-1)
        exact_domain_id = 1
        
    if exact_domain_id ==0:
        lin_space = domain[0] + np.arange(n_steps)*step_size
    else:
        lin_space = domain[1] - np.arange(n_steps)[::-1]*step_size
    new_domain = [lin_space.min(),lin_space.max()]
        
    parameters = {'step_size':step_size,'number_of__steps':n_steps,'domain':new_domain}
    if return_parameters:
        return lin_space,parameters
    else:
        return lin_space
        



import numpy as np
import numpy.ma as mp
import scipy.integrate as spIntegrate
import logging
from itertools import repeat

log=logging.getLogger('root')

import xframe.library.pythonLibrary as pyLib
import xframe.library.mathLibrary as mLib
from .harmonic_transforms import generate_circular_harmonic_transform_pair
from .harmonic_transforms import HarmonicTransform
from .hankel_transforms_zernike import generate_ht
from xframe import database
from xframe.settings import analysis as settings

doublePrecision=pyLib.doublePrecision



#####################
###  Zernike FTs  ###
def select_harmonic_transforms(harm_trf,dimensions,use_gpu):
    if (dimensions == 3) and use_gpu:
        trfs = harm_trf.transforms_by_indices['direct']
    elif (dimensions == 3) and (not use_gpu):
        trfs = harm_trf.transforms_by_indices['ml']
    elif  (dimensions == 2):
        trfs = harm_trf.transforms_by_indices['m']
    ht,iht = trfs['forward'],trfs['inverse']
    return ht,iht

def generate_zernike_ft(r_max,weights,harm_trf,dimensions,use_gpu=False,pi_in_q = False,mode = 'trapz'):
    orders=weights['posHarmOrders']
    weights_matrix = weights['weights']
    hankel,ihankel = generate_ht(weights_matrix,orders,r_max,pi_in_q=pi_in_q,dimensions=dimensions,use_gpu=use_gpu,mode=mode)
    ht,iht = select_harmonic_transforms(harm_trf,dimensions,use_gpu)
    
    def ft(data):
        harmonic_coefficients=ht(data)    
        reciprocal_harmonic_coefficients=hankel(harmonic_coefficients)
        reciprocal_function=iht(reciprocal_harmonic_coefficients)
        reciprocal_function[0] = reciprocal_function[0].mean()
        return reciprocal_function
    
    def ift(data):
        reciprocal_harmonic_coefficients=ht(data)
        harmonic_coefficients=ihankel(reciprocal_harmonic_coefficients)
        real_function=iht(harmonic_coefficients)
        real_function[0] = real_function[0].mean()
        return real_function
    
    return ft,ift




import numpy as np
import time
from xframe.settings import general
general.n_control_worker=0
import xframe
xframe.setup_experiment('fxs3046_online','normalized_proc')
from psutil import virtual_memory


db = xframe.database.experiment
#for m in range(16):
module = 3
#f0 = db.load('vds_module',as_h5_object=True,path_modifiers={'run':169,'module':3,'data_mode':'proc'})
f0_2 = db.load('/gpfs/exfel/theory_group/user/berberic/p3046/vds/r0169/vds_proc.h5',as_h5_object=True)
f0 = db.load('/gpfs/exfel/theory_group/user/berberic/p3046/vds/r0169/vds_proc_module_%d.h5'%module,as_h5_object=True)
f0_check = db.load('/gpfs/exfel/exp/SPB/202202/p003046/proc/r0169/CORR-R0169-AGIPD%02d-S00000.h5'%module,as_h5_object=True)

data_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/data'
mask_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/mask'
gain_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/gain'
train_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/trainId'
pulse_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/pulseId'
cell_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/cellId'

check_data_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/%dCH0:xtdf/image/data'%module
check_mask_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/%dCH0:xtdf/image/mask'%module
check_gain_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/%dCH0:xtdf/image/gain'%module
check_train_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/%dCH0:xtdf/image/trainId'%module
check_pulse_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/%dCH0:xtdf/image/pulseId'%module
check_cell_path = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/%dCH0:xtdf/image/cellId'%module

pulses = f0[pulse_path][:]
p_any_false = (pulses == 65535).any()
cells = f0[cell_path][:]
c_any_false = (cells == 65535).any()
trains = f0[train_path][:]
t_any_false = (trains == 65535).any()

data2 = f0_2[data_path]
data = f0[data_path]
check_data = f0_check[check_data_path]


#for m in range(16):
#    log.info(i)
#    data = f0[data_path][i::n_parts]
#    any_non = np.isnan(data).any() 
    
#c = next(exp.get_data(opt))

#c = db.load('data_chunk',frame_ids = np.arange(1,10001), **opt)
#for c_id , c in enumerate(exp.get_data(opt)):
#    del(c)
#    log.info('chunk {} done!'.format(c_id))

#start = time.time()
#c = db.load_chunks('chunks',**opt)
#log.info(time.time()-start)
#for i in range(169,170):
#    db.create_vds(i,False,create_modulewise_vds = True, create_complete_vds = False,n_processes = 40)



if False:
    def lload(m,run,N,**kwargs):
        f = db.load('vds_module',skip_custom_methods = True ,as_h5_object=True,path_modifiers={'run':run,'data_mode':'proc','module':m})['INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/data']
        out = np.zeros((N,)+f.shape[1:],dtype = f.dtype)
        f.read_direct(out,slice(100,N+100))
        return 1
    def load(run,N,**kwargs):
        f = db.load('vds',skip_custom_methods = True ,as_h5_object=True,path_modifiers={'run':run,'data_mode':'proc'})['INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/data']
        out = np.zeros((16,N)+f.shape[2:],dtype = f.dtype)
        f.read_direct(out,(slice(None),slice(100,N+100)))
        return 1
    
    runs = np.array([112,113,114,115])
    runs = np.arange(4000)
    N = 1000
    start = time.time()
    out = np.zeros((16,N)+f.shape[2:],dtype = f.dtype)
    #load(169,N)
    xframe.Multiprocessing.comm_module.request_mp_evaluation(lload,argArrays=[np.arange(16)],const_args=[169,N],n_processes=16,callWithMultipleArguments=False,assemble_outputs = False)
    stop = time.time()-start
    print(stop)



import time
import logging
import sys
import os
import numpy as np
import traceback
file_path = os.path.realpath(__file__)
plugin_dir = os.path.dirname(file_path)
os.chdir(plugin_dir)

from xframe.library import mathLibrary as mLib
from xframe.library.mathLibrary import SampleShapeFunctions,ExponentialRamp,LinearRamp
from xframe.library.gridLibrary import uniformGrid_func
from xframe.library.gridLibrary import ReGrider
from xframe.library.gridLibrary import NestedArray
from xframe.library.gridLibrary import GridFactory
from xframe.library.pythonLibrary import plugVariableIntoFunction
from xframe.library.pythonLibrary import DictNamespace
from xframe.library.pythonLibrary import RecipeFactory
from xframe.library.pythonLibrary import selectElementOfFunctionOutput_decorator
from xframe.library.pythonLibrary import xprint
from xframe.interfaces import ProjectWorkerInterface


from .projectLibrary.hankel_transforms import generate_ht
from .projectLibrary.hankel_transforms import generate_weightDict
from .projectLibrary.ft_grid_pairs import get_grid
from .projectLibrary.ft_grid_pairs import polarFTGridPair_PNAS
from .projectLibrary.ft_grid_pairs import radial_grid_func_zernike
from .projectLibrary.ft_grid_pairs import spherical_ft_grid_pair_zernike
from .projectLibrary.ft_grid_pairs import polar_ft_grid_pair_zernike
from .projectLibrary.ft_grid_pairs import polarFTGridPair_SinCos_new
from .projectLibrary.fourier_transforms import generate_ft,load_fourier_transform_weights
from .projectLibrary.misk import getAnalysisRecipeFacotry
from .projectLibrary.misk import get_analysis_process_factory
from .projectLibrary.misk import generate_calc_center
from .projectLibrary.misk import generate_absolute_value
from .projectLibrary.misk import generate_square
from .projectLibrary.misk import _get_reciprocity_coefficient
from .projectLibrary.misk import save_to_dict
from .projectLibrary.misk import copy as copy_operation
from .projectLibrary.misk import diff as diff_operation
from .projectLibrary.misk import add_above_zero_index as add_above_zero_index
from .projectLibrary.misk import load_from_dict
from .projectLibrary.fxs_Projections import ReciprocalProjection
from .projectLibrary.fxs_Projections import ShrinkWrapParts
from .projectLibrary.fxs_Projections import generate_negative_shift_operator,generate_shift_by_operator
from .projectLibrary.fxs_Projections import generate_initial_support_mask
from .projectLibrary.fxs_Projections import generate_orientation_matching
from .projectLibrary.fxs_Projections import generate_fix_point_inversion
from .projectLibrary.fxs_Projections import RealProjection
from .projectLibrary.fxs_invariant_tools import generate_estimate_number_of_particles_new,generate_estimate_number_of_particles_new_2
from .projectLibrary.fxs_IO_methods import generate_error_routines
from .projectLibrary.fxs_IO_methods import HIOProjection
from .projectLibrary.fxs_IO_methods import generate_main_error_routine
from .projectLibrary.fxs_IO_methods import error_reduction
from .projectLibrary.harmonic_transforms import HarmonicTransform
from .projectLibrary.fxs_invariant_tools import harmonic_coeff_to_deg2_invariants_2d
from .projectLibrary.fxs_invariant_tools import harmonic_coeff_to_deg2_invariants_3d
from .projectLibrary.fxs_invariant_tools import ccd_associated_legendre_matrices
from xframe.library.physicsLibrary import ewald_sphere_theta
from xframe import settings
from xframe import database
from xframe import Multiprocessing
import xframe


log=logging.getLogger('root')


opt = None
db = None
comm_module = None
def set_globals():
    global opt
    opt = settings.project
    global db
    db = database.project
    global comm_module
    comm_module = Multiprocessing.comm_module

def update_settings_and_db(new_opt):
    settings.project = new_opt
    db.update_settings(new_opt.dict())
    global opt
    opt = new_opt


class ProjectWorker(ProjectWorkerInterface):
    def __init__(self):
        #log.info('analysis worker init')
        self.opt = settings.project
        self.process_factory=RecipeFactory({})
        MTIP.preinit()
        self.mtip = MTIP
        if settings.project.GPU.use:
            n_gpus = Multiprocessing.get_number_of_gpus()
            if n_gpus == 0:
                log.warning('Problem with GPU initialization. There maybe no GPU or pyOpenCL is missing. Defaulting to CPU only computations.')
                settings.project.GPU.use = False
                settings.general.n_control_workers = 0
                xprint('Starting Reconstruction in CPU only mode. This is gonna be slow!')
            else:
                settings.general.n_control_workers = settings.project.GPU.get("n_gpu_workers",6)
            Multiprocessing.comm_module.restart_control_worker()
        else:
            xprint('Starting Reconstruction in CPU only mode. This is gonna be slow!')

        self.results={}
        self.results['stats']={}
    
    def setup_phasing_loop(self):
        self.mtip = MTIP(self.process_factory)
        self.mtip.generate_phasing_loop()
    def start_phasing(self):
        run_profiling=settings.project.get('profiling',DictNamespace(enable=False)).enable
        m = self.mtip
        #log.info("run profiling = {}".format(run_profiling))
        if run_profiling:
            process_id = settings.project.profiling.reconstruction_process_id
            if not settings.project.multi_process.use:
                process_id=0
            log.info("Profiling: id = {} profile_id = {}".format(Multiprocessing.get_process_name(),process_id))
            if Multiprocessing.get_process_name() == np.abs(process_id):
                path = db.get_reconstruction_path()
                db.create_path_if_nonexistent(path)
                path += "reconstruction_{}.stats".format(process_id)
                log.info('profile path = {}'.format(path))
                import cProfile
                profiler = cProfile.Profile()
                profiler.enable()
                result = m.phasing_loop()
                profiler.disable()
                profiler.dump_stats(path)                    
            else:
                result = m.phasing_loop()
        else:
            result = m.phasing_loop()
        return result
        
    def start_parallel_phasing(self):
        def generate_phasing_run_phasing(*opt,**kwargs):
            #xprint('Generating phasing loop.',)
            self.setup_phasing_loop()
            #xprint('done.\n')
            #xprint(f'{self.mtip.loops_str}start phasing')
            result = self.start_phasing()
            #xprint('Finished phasing!\n')
            return result
        
        mp_mode = Multiprocessing.MPMode_Queue(assemble_outputs = False)        
        mp_opt = settings.project.multi_process
        xprint(f'Spawning phasing processes executing:\n{self.mtip.loops_str}')
        n_processes = Multiprocessing._read_number_of_processes(mp_opt.n_parallel_reconstructions)
        result = Multiprocessing.comm_module.request_mp_evaluation(generate_phasing_run_phasing,input_arrays=[np.arange(n_processes)],mode=mp_mode,call_with_multiple_arguments=False,n_processes = n_processes)
        xprint('\nFinished phasing!')
        return result


    def post_processing(self):
        db = database.project
        opt = settings.project
        save = db.save
        try:
            processed_results=self.results.get('MTIP',{})
            stats=self.results.get('stats',{})
            processed_results_list=[result_dict for result_dict in processed_results.values()]
            #log.info([type(result_dict) for result_dict in processed_results])
            #log.info([result_dict.shape for result_dict in processed_results])
            errors = []
            for result_dict in processed_results_list:
                grid_pair = result_dict.pop('grid_pair')
                errors.append(result_dict['error_dict']['main'][-1])
                projection_matrices = result_dict.pop('projection_matrices')                
            r_ids=np.argsort(errors)
            log.info('error sorted reconstruction = {} \n errors ={}'.format(r_ids,np.array(errors)[r_ids]))
            processed_results_dict={str(_id):processed_results_list[_id] for _id in r_ids}
            reciprocity_coefficient = _get_reciprocity_coefficient(settings.project.fourier_transform)
            data_dict={'configuration':{'internal_grid':grid_pair,'xray_wavelength':self.mtip.load_mtip_data()[0]['xray_wavelength'],'reciprocity_coefficient':reciprocity_coefficient},'reconstruction_results':processed_results_dict,'projection_matrices':projection_matrices,'stats':stats}
            save('reconstructions',data_dict)
        except Exception as e:
            log.info(f'Error during postprocessing / saving with message:\n {e}')
            log.debug(traceback.format_exc())
            
    
    def run(self):
        #log.info(self)
        #log.info(self.mtip_opt)
        opt = settings.project
        start_time = time.time()
        #self.MTIP.preinit()
        #xprint(f'Reciprocal opt = {opt.projections.reciprocal.dict()}')
        #xprint(f'grid opt = {opt.grid.dict()}')
        if opt.multi_process.use:
            result=self.start_parallel_phasing()
        else:
            xprint('Generating phasing loop.')
            self.setup_phasing_loop()
            xprint('done.\n')
            xprint(f'Start phasing:\n{self.mtip.loops_str} ')
            data = self.start_phasing()
            xprint('\nFinished phasing!')
            result = np.array([data],dtype = object)
            
        xprint('\nSaving results.')
        self.results['MTIP']=result
        self.results['stats']['run_time'] = time.time()-start_time
        self.post_processing()
        return result,locals()
    
class MTIP:
    # all subsequent parameters are set by a call to MTIP.preinit
    dimensions = 'not set'
    mtip_data = 'not_set'
    data_q_limits = 'not_set'
    
    data_number_of_radial_points = 'not_set'
    real_radial_points = 'not_set'
    reciprocal_radial_points = 'not_set'
    max_q = 'not_set'
    fourier_transform_weights = 'not_set'
    error_metric_parameters_dict = {} 
    preinit_was_called = False

    loops_str = 'Loops:\n'
    loops_opt = settings.project.main_loop.sub_loops
    max_loop_name_size = max([len(name)for name in loops_opt.order])
    for loop_name in loops_opt.order:            
        loop_opt = loops_opt.get(loop_name,{})
        iterations = loop_opt.get('iterations','')
        methods_string = ''
        for method in loop_opt.get('order',[]):
            method_opt = loop_opt.methods.get(method,{})
            if isinstance(method_opt,(dict,DictNamespace)):
                method_iteration = method_opt['iterations']
            else:
                method_iteration = method_opt
            methods_string += f'{method_iteration}x{method} '
        loops_str += f'\t{loop_name}:'+' '*(max_loop_name_size-len(loop_name))+ f'\t {iterations}x( '+''.join(methods_string)+')\n'
   
    @classmethod
    def preinit(cls):
        ''' 
        This Method prepares needed quantities which require multiprocessing that are equal for all MTIP instances before __init__ is called in a multi process environment. This is currently done only to preload the fourier transform weights.
        '''
        opt = settings.project
        cls.dimensions = opt.dimensions
        tmp = cls.load_mtip_data()
        cls.mtip_data = tmp[0]
        cls.data_q_limits = tmp[1]
        cls.data_number_of_radial_points = tmp[2]

        # create fourier transform weights
        ht_opt={
            'dimensions':opt.dimensions,
            **opt.grid,            
        }
        max_q = opt.grid.max_q
        if not isinstance(max_q,float):
            max_q = cls.data_q_limits[1]
        cls.max_q=max_q
        # generate mock angular grid parameters to use the standard grid routine to get the internal radial grid.
        if opt.dimensions ==2:
            grid_param = {'phis':np.array([1.0,2.0])}
        elif opt.dimensions ==3:
            grid_param = {'phis':np.array([1.0,2.0]),'thetas':np.array([1.0,2.0])}
        mock_grid_pair = get_grid({**opt.fourier_transform,**ht_opt,**grid_param,'max_q':max_q,'n_radial_points_from_data':cls.data_number_of_radial_points})        
        qs = mock_grid_pair.reciprocalGrid.__getitem__((slice(None),)+(0,)*cls.dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
        cls.reciprocal_radial_points = qs
        rs = mock_grid_pair.realGrid.__getitem__((slice(None),)+(0,)*cls.dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
        cls.real_radial_points = rs        
        n_radial_points = len(qs)
        r_max = np.max(rs)
        #cls.fourier_transform_weights = cls.load_fourier_transform_weights(n_radial_points,r_max)
        cls.fourier_transform_weights = load_fourier_transform_weights(opt.dimensions,opt.fourier_transform,opt.grid,database.project)
        cls.preinit_was_called = True
    @classmethod
    def load_mtip_data(cls):
        opt = settings.project
        db = database.project
        data = db.load('invariants',path_modifiers={'structure_name':opt.structure_name,'dimensions':opt.dimensions})        
        data_q_limits = [data['data_radial_points'].min(),data['data_radial_points'].max()]
        data_number_of_radial_points = len(data['data_radial_points'])
        #log.info('data points = {}'.format(data['data_radial_points']))
        return [data,data_q_limits,data_number_of_radial_points]
    

    def __init__(self,process_factory):
        set_globals()
        #self.opt = settings.project
        self.process_factory=process_factory
        self.results={}
        
        # will be set by generate_phasing_loop
        self.grid_pair = False
        self.routines = False
        self.projection_objects=False
        self.mtip_part_names = False
        self.mtip_io_names = False
        self.phasing_loop = False        


    def assemble_operators(self):
        #order is important, e.g. assemble_error_routines depents on self.projection_objects
        transforms,grid_pair,max_order=self.assemble_transform_op_and_grid()        
        reciprocal_projection,rp_obj=self.assemble_reciprocal_projection_op(grid_pair,max_order)
        real_projections,hio_projection,real_projection_obj,shrink_wrap_obj=self.assemble_real_projection_op(grid_pair,rp_obj,transforms)
        self.projection_objects = {'reciprocal':rp_obj,'hio':hio_projection,'real':real_projection_obj,'sw':shrink_wrap_obj}
        orientation_fixing=self.assemble_orientation_fixing_op(grid_pair)
        errors = self.assemble_error_routines(grid_pair)
        misk = self.assemble_misk_oppesators(grid_pair)


        #log.info(f'{transforms}\n {reciprocal_projection}\n{real_projections}\n{orientation_fixing}\n{errors}\n {misk}')
        operators={**transforms,**real_projections,**reciprocal_projection,**orientation_fixing,**errors,**misk}
        return operators,grid_pair
    
    def generate_fourier_transforms(self,harm_trf):
        dimensions = self.dimensions
        ft_opt = opt.fourier_transform
        ft_type = ft_opt['type']
        pos_orders= np.arange(harm_trf.max_order+1)
        pi_in_q = ft_opt.get('pi_in_q',None)

        reciprocity_coefficient=_get_reciprocity_coefficient(ft_opt)
        
        use_gpu = opt['GPU']['use']                    
        weights_dict = self.fourier_transform_weights
        r_max = np.max(self.real_radial_points)
        #log.info('Use GPU = {}'.format(use_gpu))
        fourierTransform,inverseFourierTransform=generate_ft(r_max,weights_dict,harm_trf,dimensions,pos_orders=pos_orders,reciprocity_coefficient=reciprocity_coefficient,use_gpu = use_gpu,mode = ft_type)
        
        #d= np.ones((settings.project.grid.n_radial_points,pos_orders.max()*2+1))
        #d2 = inverseFourierTransform(fourierTransform(d))
        #log.info(d2[:,0])
        return fourierTransform,inverseFourierTransform

    def assemble_transform_op_and_grid(self):
        ht_opt={
            'dimensions':opt.dimensions,
            **opt.grid            
        }
        #log.info(f'ht opt = {ht_opt}')
        # Create harmonic transforms (needs to happen before grid selection since harmonic transform plugin can choose the angular part of the grid.)
        cht=HarmonicTransform('complex',ht_opt)
        #ht_data_type = opt['harmonic_transform']['data_type']
        if self.dimensions == 2:
            ht =  HarmonicTransform('real',ht_opt)
        elif self.dimensions == 3:
            ht=cht
        cht_forward,cht_inverse = cht.forward, cht.inverse
        #log.info(cht_forward)
        #log.info(f'spat shape = {cht._sh._sh.spat_shape} ht_opt = {ht_opt}')
        ht_forward,ht_inverse = ht.forward, ht.inverse
        #log.info(ht_forward)
        #xprint("ht grid param = {}".format(cht.grid_param))
        #log.info(f'spat shape = {cht._sh._sh.spat_shape}')
        grid_pair=get_grid({**opt.fourier_transform,**ht_opt,**cht.grid_param,'max_q':self.max_q,'n_radial_points_from_data':self.data_number_of_radial_points})
        # fourier transforms
        ft_forward,ft_inverse=self.generate_fourier_transforms(cht)
        #log.info(f'grid shape = {grid_pair.realGrid[:].shape}')
        #log.info("test ft:")
        #a = np.ones(grid_pair.realGrid.shape)
        #a[20:]=0
        #for i in range(20):
        #    a=ft_inverse(ft_forward(a))
        #log.info('mean at 10 = {} mian at 40 = {}'.format(np.mean(a[10]),np.mean(a[40])))
        #raise Exception()
        
        transform_op={'fourier_transform':ft_forward,'inverse_fourier_transform':ft_inverse,'harmonic_transform':ht_forward,'inverse_harmonic_transform':ht_inverse,'complex_harmonic_transform':cht_forward,'complex_inverse_harmonic_transform':cht_inverse}
        return transform_op,grid_pair,ht.max_order

    def assemble_reciprocal_projection_op(self,grid_pair,max_order):
        #opt=opt['projections']['reciprocal']
        rp=ReciprocalProjection(grid_pair.reciprocalGrid,self.mtip_data,max_order)
        self.rprojection = rp
        self.results['n_particles']=[]
        self.results['n_particles_fraction']=[]
        self.results['n_particles_gradients']=[]        
        def save_number_of_particles():
            n_dict = rp.number_of_particles_dict
            #log.info(n_dict['number_of_particles'])
            self.results['n_particles'].append(n_dict['number_of_particles'])
            neg_fractions = n_dict['negative_fraction']
            gradient = n_dict['gradient']
            if isinstance(neg_fractions,np.ndarray):
                self.results['n_particles_fraction'].append(neg_fractions)
            if isinstance(gradient,np.ndarray):
                self.results['n_particles_gradients'].append(gradient)
                
        reciprocal_projection_op={'mtip_projection':rp.mtip_projection,'approximate_unknowns':rp.approximate_unknowns,'project_to_modified_intensity':rp.project_to_modified_intensity,'project_to_fixed_intensity':rp.project_to_fixed_intensity,'save_number_of_particles':save_number_of_particles}
        return reciprocal_projection_op,rp
    
    def assemble_real_projection_op(self,grid_pair,rp_obj,transforms):
        #log.info('assemble real projection')
        r_opt=opt.projections.real
        r_grid=grid_pair.realGrid
        q_grid=grid_pair.reciprocalGrid

        # create autocorrelation
        pr = rp_obj.full_projection_matrices
        icht = transforms['inverse_harmonic_transform']
        ift = transforms['inverse_fourier_transform']
        #log.info(f'len pr = {len(rp_obj.used_orders)}')
        #log.info(f'pr shapes = {[p.shape for p in pr]}')
        #log.info(f'len pr = {len(pr)}')

        if self.dimensions ==3:
            pr_padded = []
            for l,p in enumerate(pr):
                n_ms = 2*l+1
                if p.shape[1]!= n_ms:
                    p_padded = np.zeros((p.shape[0],n_ms),dtype=p.dtype)
                    p_padded[:,:p.shape[1]]=p
                    pr_padded.append(p_padded)
                else:
                    pr_padded.append(p)
                        
            #xprint(f'pr shape = {[p.shape for p in pr_padded]}')
            auto_correlation_guess = ift(icht(pr_padded)).real
        elif self.dimensions ==2:
            #log.info(f'pr.shape = {pr.shape}')
            auto_correlation_guess = ift(icht(pr.T)).real
        
        metadata={'integrated_intensity':rp_obj.integrated_intensity,'real_grid':r_grid,'auto_correlation':auto_correlation_guess}

        real_projection_obj = RealProjection(r_opt.projections,metadata)

        sw_opt=r_opt['shrink_wrap']
        sw_obj = ShrinkWrapParts(r_grid,q_grid,real_projection_obj.initial_support)
        #real_support_projection=generate_real_support_projection(initial_support_mask)
        
        #other_real_projections=generate_other_real_projections(r_opt.non_support_projections,metadata)
    
        
        multiply_ft_gaussian=sw_obj.multiply_with_ft_gaussian  #generateSW_multiply_ft_gaussian(sw_opt['sigma'],q_grid)
        calculate_support_mask=sw_obj.get_new_mask
        #hio = generate_HIO(r_opt['HIO'])
        hio_opt = r_opt.HIO
        hio = HIOProjection(hio_opt.beta[0],considered_projections=hio_opt.get('considered_projections',['all']))
        #hio = generate_HIO({'beta':self.beta})
        
        #real_projection_op={'real_support_projection':real_support_projection,'other_real_projections':other_real_projections,'multiply_ft_gaussian':multiply_ft_gaussian,'calculate_support_mask':calculate_support_mask,'hybrid_input_output':hio.projection,'real_projection':real_projection_obj.projection}
        
        real_projection_op={'multiply_ft_gaussian':multiply_ft_gaussian,'calculate_support_mask':calculate_support_mask,'hybrid_input_output':hio.projection,'error_reduction':error_reduction,'real_projection':real_projection_obj.projection}
        
        return real_projection_op,hio,real_projection_obj,sw_obj
    
    def assemble_orientation_fixing_op(self,grid_pair):
        ft_opt=opt['fourier_transform']
        r_grid=grid_pair.realGrid
        q_grid=grid_pair.reciprocalGrid
        calc_center_routine=generate_calc_center(r_grid)
        negative_shift_routine = generate_shift_by_operator(grid_pair.reciprocalGrid,opposite_direction = True)
        #negative_shift_routine=generate_negative_shift_operator(q_grid,ft_opt['type'])
        #negative_shift_routine=generate_negative_shift_operator(q_grid,ft_opt['type'])
        orientation_fixing_op={'calc_center':calc_center_routine,'negative_shift':negative_shift_routine}
        if opt.dimensions==2:
            if opt.output_density_modifiers.fix_orientation:
                remaining_SO_projection=self.projection_objects['reciprocal'].remaining_SO_projection
                orientation_fixing_op['fix_remaining_SO_freedom']=remaining_SO_projection
        return orientation_fixing_op

    def assemble_error_routines(self,grid_pair):
        error_opt = opt.main_loop.error
        rp = self.projection_objects['reciprocal']
        deg2_invariants = rp.deg2_invariants
        number_of_particles = rp.number_of_particles
        used_orders = rp.used_orders
        xray_wavelength = self.mtip_data['xray_wavelength']
        invariant_mask = rp.radial_mask[:,:,None] * rp.radial_mask[:,None,:]
        initial_mask = self.projection_objects['real'].initial_support
        error_routines = generate_error_routines(error_opt,grid_pair,deg2_invariants = deg2_invariants,projection_matrices=rp.projection_matrices,used_orders = used_orders,n_particles = rp.number_of_particles,invariant_mask = invariant_mask,xray_wavelength = xray_wavelength,initial_mask = initial_mask)
        #log.info('error_routines {}'.format(error_routines))
        return {'real_errors':error_routines['real'],'reciprocal_errors':error_routines['reciprocal']}
        
    def assemble_misk_oppesators(self,grid_pair):
        dtype = np.dtype(complex)
        abs_value = generate_absolute_value(grid_pair.realGrid[:].shape[:-1],dtype,cache_aware = settings.general.cache_aware,L2_cache = settings.general.L2_cache)        
        square = generate_square(grid_pair.realGrid[:].shape[:-1],dtype,cache_aware = settings.general.cache_aware,L2_cache = settings.general.L2_cache)
        if self.dimensions == 3:
            calc_deg2_invariant = harmonic_coeff_to_deg2_invariants_3d
        elif self.dimensions == 2:
            calc_deg2_invariant = harmonic_coeff_to_deg2_invariants_2d
        return {'square_grid':square,'abs_value':abs_value,'calc_deg2_invariant':calc_deg2_invariant,'save_to_dict':save_to_dict,'copy':copy_operation,'add_above_zero_index':add_above_zero_index,'diff':diff_operation,'load_from_dict':load_from_dict}
        
    
    def assemble_MTIP_routines(self):
        error_saving_parts = self.assemble_error_saving_parts()
        self.process_factory.addOperators(error_saving_parts)
        mtip_parts=self.assemble_mtip_parts()
        self.process_factory.addOperators(mtip_parts)
        io_variants=self.assemble_IO_variants()
        output_modifier=self.assemble_output_modifier(opt['output_density_modifiers'])
        calc_deg2_invariant = self.assemble_calc_deg2_invariant()
        #log.info(io_variants['SW'])
        return {'calc_deg2_invariant':calc_deg2_invariant,'output_modifier':output_modifier,**io_variants}

    def assemble_error_saving_parts(self):
        error_parts = {}
        process_factory = self.process_factory
        for name in opt.main_loop.error.methods:
            if name!='main':
                error_sketch = [name+'_errors',
                                [(0,),[('save_to_dict',(self.results,['errors',name],'iterative_append'))]]
                                ]
                try:
                    error_parts['calc_' + name + '_errors'] = process_factory.buildProcessFromSketch(error_sketch)
                except KeyError as e:
                    error_sketch = [(0,),['id']]
                    error_parts['calc_' + name +'_errors'] = process_factory.buildProcessFromSketch(error_sketch)
        
        return error_parts
            
    def assemble_mtip_parts(self):
        process_factory=self.process_factory
        routine_sketches = {}
        routine_sketches['MTIP_start']=[
            [(0,0),['copy','square_grid']],                           
            [(0,1,1),['id','harmonic_transform','copy']],
            [(0,1,1,2),['id','id','approximate_unknowns','id']],
            [(0,1,2,3),[('id',()),('id',()),('save_to_dict',(self.results,'fxs_unknowns','replace')),'id']],
            [(0,1,2,1,3),['id','mtip_projection','id','id']],
            [(0,1,2,3),['id','inverse_harmonic_transform','id','id']],
            [(0,0,3,1,2),['id','project_to_modified_intensity','save_number_of_particles','id']],
            [(0,1,2,1),['calc_reciprocal_errors','id']],
            [(1,),['id']]
        ]

        routine_sketches['MTIP_start_non_FXS']=[
            [(0,0),['copy','square_grid']],                           
            [(0,0,1,1),['id','project_to_fixed_intensity','save_number_of_particles','id']],
            [(0,1,2,1),['calc_reciprocal_errors','id']],
            [(1,),['id']]
        ]             

        #routine_sketches['MTIP_start_estimate_n_particles']=[
        #    [(0,0),['copy','square_grid']],                           
        #    [(0,1),['id','harmonic_transform']],
        #    [(0,1,1),['id','id','approximate_unknowns']],
        #    [(0,1,2),[('id',()),('id',()),('save_to_dict',(self.results,'fxs_unknowns','replace'))]],
        #    [(0,1,2,1),['id','mtip_projection','id']],
        #    [(0,1,2),['id','inverse_harmonic_transform','id']],
        #    [(0,1,2),['id','particle_number_projection','id']],
        #    [(0,0,1,2),['id','project_to_modified_intensity','save_number_of_particles','id']],
        #    [(0,1,2,1),['calc_reciprocal_errors','id']],
        #    [(1,),['id']]
        #]
        
        routine_sketches['MTIP_end']=['id',
                         ['id','copy'],
                         ['fourier_transform','id']
                        ]
        routines = {key:process_factory.buildProcessFromSketch(sketch) for key,sketch in routine_sketches.items()}
        self.mtip_part_names = tuple(routines.keys())
        return routines
        
    def assemble_IO_variants(self):
        estimate_number_of_particles = settings.project.projections.reciprocal.number_of_particles.get('estimate',False)
        estimate_in_variants = settings.project.projections.reciprocal.number_of_particles.get('estimate_in',[])
        #use_ht_difference = settings.project.main_loop.get('stabilize_harmonic_transform',False)       
        #log.info('using ft difference = {}'.format(use_ft_difference))
        #log.info('using ht difference = {}'.format(use_ht_difference))
        start_routine_name_initial='MTIP_start'
        routine_names = ('HIO','ER','HIO_non_FXS','ER_non_FXS')
        io_names={'HIO':'hybrid_input_output','ER':'error_reduction','HIO_non_FXS':'hybrid_input_output','ER_non_FXS':'error_reduction'}      
            
        routine_sketches={}
        for name in routine_names:
            start_routine_name=start_routine_name_initial
            if '_non_FXS' in name:
                start_routine_name+='_non_FXS'
            elif (name in estimate_in_variants) and estimate_number_of_particles:
                start_routine_name+='_estimate_n_particles'

            sketch = [
                [(1,1),['fourier_transform','id']],
                [np.array([0,0,1],dtype=int),[start_routine_name,'id']],
                [np.array([0,1,0],dtype=int),['inverse_fourier_transform','id','id']],
                [np.array([0,0,1,2],dtype=int),['copy','real_projection','id','id']],
                [(0,1,2,0,1,3),[io_names[name],'calc_real_errors','id']],
                [(2,0),['id','id']]
            ]
            sketch_ft_stab = [
                    [(1,1),['fourier_transform','id']],
                    [(0,0,0,1),[start_routine_name,'inverse_fourier_transform','id']],
                    [(0,2,1,2,0),['inverse_fourier_transform','diff','id','id']],
                    [(0,1,2,3),['add_above_zero_index','id','id']],
                    [(0,0,1,2),['copy','real_projection','id','id']],
                    [(0,1,2,0,1,3),[io_names[name],'calc_real_errors','id']],
                    [(2,0),['id','id']]
                    #[(0,),['MTIP_end']]                    
                ]
            
            routine_sketches[name]=sketch
            routine_sketches[name+'_ft_stab']=sketch_ft_stab

        routine_sketches['SW']=[
            'copy',
            ['abs_value','copy'],
            [(0,1),['fourier_transform','id']],
            [(0,1),['multiply_ft_gaussian','id']],
            [(0,1),['inverse_fourier_transform','id']],
            [(0,1),['calculate_support_mask']]
        ]
        routine_sketches['SW_center']=[
            [(0,0,0),['copy','fourier_transform','copy']],
            [(0,0,1,2),['abs_value','copy','id','id']],
            [(0,1,2,3),['fourier_transform','id','id','id']],
            [(0,1,2,3),['multiply_ft_gaussian','id','id','id']],
            [(0,1,2,3),['inverse_fourier_transform','id','id','id']],
            [(0,1,2,3),['calculate_support_mask','id','id']]
        ]
        self.mtip_io_names = tuple(routine_sketches.keys())
        #log.info(f'routine IO names = {self.mtip_io_names}')
        process_factory=self.process_factory
        routines = {key:process_factory.buildProcessFromSketch(sketch) for key,sketch in routine_sketches.items()}
        #log.info(f'routine_names = {routines.keys()}')
        return routines

    def assemble_mtip_parts_old(self):
        process_factory=self.process_factory

        #standard working phasing part
        mtip_start_sketch=['id',
                           ['copy','id'],                           
                           [(0,1),['id','square_grid']],                           
                           [(0,1),['id','harmonic_transform']],
                           [(0,1,1),['id','id','approximate_unknowns']],
                           [(0,1,2),[('id',()),('id',()),('save_to_dict',(self.results,'fxs_unknowns','replace'))]],
                           [(0,1,2,1),['id','mtip_projection','id']],
                           [(0,1,2),['id','inverse_harmonic_transform','id']],
                           [(0,0,1,2),['id','project_to_modified_intensity','save_number_of_particles','id']],
                           [(0,1,2,1),['calc_reciprocal_errors','id']],
                           [(1,),['inverse_fourier_transform']]
                           ]
        
        #mtip_start_estimate_n_particles_sketch=['id',
        #                                        ['copy','id'],                           
        #                                        [(0,1),['id','square_grid']],                           
        #                                        [(0,1),['id','harmonic_transform']],
        #                                        [(0,1,1),['id','id','approximate_unknowns']],
        #                                        [(0,1,2),[('id',()),('id',()),('save_to_dict',(self.results,'fxs_unknowns','replace'))]],
        #                                        [(0,1,2,1),['id','mtip_projection','id']],
        #                                        [(0,1,2),['id','inverse_harmonic_transform','id']],
        #                                        [(0,1,2),['id','particle_number_projection','id']],
        #                                        [(0,0,1,2),['id','project_to_modified_intensity','save_number_of_particles','id']],
        #                                        [(0,1,2,1),['calc_reciprocal_errors','id']],
        #                                        [(1,),['inverse_fourier_transform']]
        #                                        ]
        
        mtip_end_sketch=['id',
                         ['id','copy'],
                         ['fourier_transform','id']
                         ]

        mtip_start=process_factory.buildProcessFromSketch(mtip_start_sketch)
        mtip_start_estimate_n_particles = process_factory.buildProcessFromSketch(mtip_start_estimate_n_particles_sketch)
        mtip_end=process_factory.buildProcessFromSketch(mtip_end_sketch)        
        #mtip_start=selectElementOfFunctionOutput_decorator(mtip_start.run,0)
        
        return {'MTIP_start':mtip_start,'MTIP_end':mtip_end.run,'MTIP_start_estimate_n_particles':mtip_start_estimate_n_particles}
        
    def assemble_IO_variants_old(self):
        estimate_number_of_particles = settings.project.projections.reciprocal.number_of_particles.estimate
        estimate_in_variants = settings.project.projections.reciprocal.number_of_particles.estimate_in
        if estimate_number_of_particles and ('HIO' in estimate_in_variants):
            fxs_HIO_sketch=[
                [np.array([0,1],dtype=int),['MTIP_start_estimate_n_particles','id']],
                [np.array([0,0,1]),['copy','id','id']],       
                [np.array([0,1,2],dtype=int),['id','real_support_projection','id']],
                [np.array([0,1,2],dtype=int),['id','other_real_projections','id']],
                [(0,1,2,0,1),['hybrid_input_output','calc_real_errors']],
                [(0,),['MTIP_end']]
            ]
        else:
            fxs_HIO_sketch=[
                [np.array([0,1],dtype=int),['MTIP_start','id']],
                [np.array([0,0,1]),['copy','id','id']],       
                [np.array([0,1,2],dtype=int),['id','real_support_projection','id']],
                [np.array([0,1,2],dtype=int),['id','other_real_projections','id']],
                [(0,1,2,0,1),['hybrid_input_output','calc_real_errors']],
                [(0,),['MTIP_end']]
            ]


        if estimate_number_of_particles and ('ER' in estimate_in_variants):
            fxs_ER_sketch=[
                [(0,1),['MTIP_start_estimate_n_particles','id']],
                [(0,0),['copy','id']],
                [(0,1),['id','real_support_projection']],
                [(0,1),['id','other_real_projections']],
                [(0,1,0,1),['error_reduction','calc_real_errors']],
                [(0,),['MTIP_end']]
            ]
        else:
            fxs_ER_sketch=[
                [(0,1),['MTIP_start','id']],
                [(0,0),['copy','id']],
                [(0,1),['id','real_support_projection']],
                [(0,1),['id','other_real_projections']],
                [(0,1,0,1),['error_reduction','calc_real_errors']],
                [(0,),['MTIP_end']]
            ]
            
        SW_sketch=[
            'copy',
            ['abs_value','copy'],
            [(0,1),['fourier_transform','id']],
            [(0,1),['multiply_ft_gaussian','id']],
            [(0,1),['inverse_fourier_transform','id']],
            [(0,1),['calculate_support_mask']]
        ]

        process_factory=self.process_factory
        fxs_HIO=process_factory.buildProcessFromSketch(fxs_HIO_sketch)
        fxs_ER=process_factory.buildProcessFromSketch(fxs_ER_sketch)
        fxs_SW=process_factory.buildProcessFromSketch(SW_sketch)
        return {'HIO':fxs_HIO,'ER':fxs_ER,'SW':fxs_SW}

    def assemble_output_modifier(self,opt):

        output_modifier_sketch=[
            [(0,1,1),['id','fourier_transform','id']],                        
        ]
        _id = [['id','id']]

        shift_center = [
            [(0,1,1),['copy','fourier_transform','calc_center']],
            [(0,1,2),[['id',np.array([],dtype=object)],['id',np.array([],dtype=object)],['save_to_dict',np.array([self.results,'neg_center_pos','replace'],dtype=object)]]],
            [(0,2,1,2),['negative_shift','negative_shift']],
            [(0,1),['id','inverse_fourier_transform']]
            #[(0,),['id']]
        ]
        
        fix_orientation=[
            [(0,1),['complex_harmonic_transform','complex_harmonic_transform']],
            [(0,1),[['id',np.array([],dtype=object)],['id',np.array([],dtype=object)],['load_from_dict',np.array([self.results,'fxs_unknowns'],dtype=object)]]],
            [(0,2,1,2),['fix_remaining_SO_freedom','fix_remaining_SO_freedom']],
            [(0,1),['complex_inverse_harmonic_transform','complex_inverse_harmonic_transform']]
        ]


        output_modifier_sketch=_id
        if opt['shift_to_center']:
            output_modifier_sketch = shift_center
        if opt.get('fix_orientation',False):
            try:
                assert settings.project.projections.reciprocal.SO_freedom.use,'SO Freedom is not used in reciprocal projection (projections.reciprocal.SO_freedom.use = False) therfore no orientation fixing is possible. skipping its assembly.'
                assert self.dimensions == 2,'3D version of MTIP does not allow for orientation fixing (only in 2D this is possible). Skipping orientation fixing assembly.'
                output_modifier_sketch = shift_center+fix_orientation
            except AssertionError as e:
                log.info(e)
        output_modifier = self.process_factory.buildProcessFromSketch(output_modifier_sketch)        
        return output_modifier

    def assemble_calc_deg2_invariant(self):
        calc_deg2_invariant_sketch = [
            ['fourier_transform'],
            ['square_grid'],
            ['harmonic_transform'],
            ['calc_deg2_invariant']
        ]
        calc_deg2_invariant = self.process_factory.buildProcessFromSketch(calc_deg2_invariant_sketch)
        return calc_deg2_invariant


    def assemble_phasing_loop(self):
        #### get relevant options ####
        sw_opt = opt.projections.real.shrink_wrap
        supp_opt = opt.projections.real.projections.support
        hio_opt = opt.projections.real.HIO
        enforce_initial_support_opt = opt.projections.real.projections.support.enforce_initial_support
        loop_opt=opt['main_loop']

        #### get routines ####
        routines=self.routines
        fourier_transform=self.process_factory.operatorDict['fourier_transform']
        inverse_fourier_transform=self.process_factory.operatorDict['inverse_fourier_transform']
        inverse_harmonic_transform=self.process_factory.operatorDict['inverse_harmonic_transform']
        prm = self.projection_objects['reciprocal'].projection_matrices
        #low_res_coeff = self.projection_objects['reciprocal'].projection_matrices
        #tmp_coeff = [np.zeros(p.shape) for p in prm]
        #for i,Ilm in enumerate(low_res_coeff):
        #    tmp_coeff[i]=Ilm
        #I = inverse_harmonic_transform(tmp_coeff)
        #I = I.real
        #I[I<0]=0
        #rho_guess = inverse_fourier_transform(I.astype(complex))
        #real_grid=self.grid_pair.realGrid
        #reciprocal_grid=self.grid_pair.reciprocalGrid
        #db.save("/gpfs/exfel/theory_group/user/berberic/MTIP/test/inital_density/d1_low_res.vts",[np.abs(rho_guess),rho_guess.real],grid_type='spherical',grid = real_grid[:])
        #db.save("/gpfs/exfel/theory_group/user/berberic/MTIP/test/inital_density/Intensity_guess.vts",[np.abs(I),rho_guess.real],grid_type='spherical',grid = reciprocal_grid[:])
        real_grid=self.grid_pair.realGrid
        real_density_guess_method=self.generate_density_guess_method(opt['density_guess'],real_grid)
        
        #### generate main error routines ####
        main_error_names = opt.main_loop.error.methods.main.metrics
        main_error_routine = generate_main_error_routine(main_error_names,opt.main_loop.error.methods.main.type)
        #check_error_target,check_error_gain_limit = self.generate_check_error_routines()
        init_error_dict = self.init_error_dict
        update_shrink_wrap = self.generate_update_shrink_wrap(self.projection_objects['sw'])
        #### assemble loops ####
        def read_method_settings(method_name,method_opt):
            if isinstance(method_opt,(dict,DictNamespace)):
                iterations = method_opt.get('iterations',0)
                out_dict = {'process':routines[method_name],'iterations':iterations,'options':method_opt}
                    
            else:
                out_dict = {'process':routines[method_name],'iterations':method_opt,'options':{}}
            return out_dict
            
            
        def generate_loop_method(loop_name,loop_opt,loop_number):
            max_iterations = loop_opt.iterations
            order = loop_opt.order
                
            methods = {key:read_method_settings(key,loop_opt['methods'][key]) for key in order}
            # HIO parameter #
            n_hio_betas = len(hio_opt.beta)
            if n_hio_betas-1 < loop_number:
                hio_beta = [0.5,0.5,-1/700,1600]
                log.info(f'No HIO beta specified in settings for loop with id {loop_number} in {loop_opt.order}. Setting beta = {hio_beta}')
            else:
                hio_beta = hio_opt.beta[loop_number]                
            hio_beta_ramp = ExponentialRamp(*hio_beta)
    
            #log.info(f'{loop_name} : threshold = {sw_threshold} from thresholds = {sw_opt.thresholds}')
            # support #
            enforce_initial_support_opt = supp_opt.enforce_initial_support
            if enforce_initial_support_opt.apply:
                enforce_initial_support_error_limit=[enforce_initial_support_opt['if_error_bigger_than']]
            else:
                enforce_initial_support_error_limit=[np.inf]

            def change_to_ft_stab(process_opt,process_name,enforce_initial_support_list):
                apply_ft_stabilization = False
                process_not_ft_stabilized = (process_name[-8:]!='_ft_stab')
                if process_not_ft_stabilized:
                    if 'ft_stab' in (process_opt):
                        if isinstance(process_opt.ft_stab,bool):
                            apply_ft_stabilization = process_opt.ft_stab
                        elif process_opt.ft_stab == 'link_to_enforce_initial_support':
                            delay = max(int(process_opt.link_to_enforce_initial_support.delay),1)
                            if len(enforce_initial_support_list)>=delay:
                                enforced_support_in_last_iterations = (np.array(enforce_initial_support_list[-delay:])==True).any()
                                #log.info(f'support_enforced_list = {enforce_initial_support_list}')
                                apply_ft_stabilization = not enforced_support_in_last_iterations
                #log.info(f'apply_ft_stabilization = {apply_ft_stabilization}')
                return apply_ft_stabilization
                            
                    
                
            def loop(state):
                if 'SW' in methods:
                    update_shrink_wrap(0,loop_number)                
                error_dict = state['error_dict']
                #log.info(f'initial error dict = \n {error_dict}')
                hist = state['density_pair_history']
                enforce_initial_support_list = state.get('enforce_initial_support_list',[])
                #log.info(enforce_initial_support_list)
                iteration = 0
                step = 0
                real_pr =  self.projection_objects['real']
                copy = np.array
                latest_intensity = False
                sw_step = 0
                for iteration in range(1,max_iterations+1):
                    #error_target_is_reached = check_error_target(error_dict)
                    #if error_target_is_reached:
                    #    break
                    for key in order:                    
                        process=methods[key]['process']
                        repeats=methods[key]['iterations']
                        process_opt = methods[key].get('options',{})
                        #log.info('Loop:{} Running {} steps of {} with error_limit {}:'.format(iteration,repeats,key,relative_error_limit))
                        if key=='SW':
                            support = process.run(state['density_pair_history'][-1][1])
                            enforce_initial_support = error_dict['main'][-1:]>enforce_initial_support_error_limit
                            enforce_initial_support_list.append(enforce_initial_support)
                            real_pr.enforce_initial_support = enforce_initial_support
                            real_pr.support = support
                            state['mask'] = real_pr.support #mask #mask
                            sw_step+=1
                            update_shrink_wrap(sw_step,loop_number)
                        elif key == 'SW_center':
                            enforce_initial_support = error_dict['main'][-1]>enforce_initial_support_error_limit
                            real_pr.enforce_initial_support = enforce_initial_support
                            enforce_initial_support_list.append(enforce_initial_support)
                            for i in range(repeats):
                                #log.info('apply SW center')
                                support,ft_density,density = process.run(state['density_pair_history'][-1][1])
                                real_pr.support = support
                                state['mask'] = real_pr.support #mask #mask
                                state['density_pair_history']=hist[1:]+((ft_density,density),)
                                sw_step+=1
                                update_shrink_wrap(sw_step,loop_number)
                        else:
                            if key in ['ER_non_FXS','HIO_non_FXS']:
                                if isinstance(latest_intensity,bool):
                                    latest_intensity = np.abs(hist[-1][0]).real
                                    self.projection_objects['reciprocal'].fixed_intensity=latest_intensity
                            else:
                                latest_intensity = False                        
                                
                            if change_to_ft_stab(process_opt,key,enforce_initial_support_list):
                                #log.info('\n\n changing to ft stab \n')
                                process = routines[key+'_ft_stab']
                            errs={}
                            for i in range(repeats):
                                self.projection_objects['hio'].beta = hio_beta_ramp.eval(step)
                                #log.info(f'beta value = {self.projection_objects["hio"].beta}')
                                hist = state['density_pair_history']
                                # Break loop if relative error limits are breached. Start new loop with best known reconstruction.
                                #if i>2:
                                #    rel_limit_is_reached = check_error_gain_limit(error_dict,key)
                                #    if rel_limit_is_reached:
                                #        #log.info('break {} loop.'.format(key))
                                #        if len(hist)>2:
                                #            state['density_pair_history'] = state['best_density_pair']
                                #        break                            
                                #log.info('current density shapes reciprocal={} real={}'.format(density_pairs[-1][0].shape,density_pairs[-1][1].shape))
                                        
                                new_density_pair = process.run(*hist[-1])                            
                                copied_density_pair = tuple(copy(a) for a in new_density_pair)
                                state['density_pair_history']=hist[1:]+(copied_density_pair,)
                                main_error = main_error_routine(error_dict)
                                #if key == 'ER':
                                #    if error_dict['main'][-1] <= 2e-3:
                                #        #methods['ER']['process'] = routines['ER_ft_stab']
                                #        pass
                                error_dict['main'].append(main_error)
                                
                                if  state['best_error'] > main_error:                                
                                    state['best_error'] = main_error
                                    state['best_density_pair'] = copied_density_pair
                                    state['best_iteration'] = iteration
                                    state['best_mask'] = state['mask']                                
                                step+=1                                
                            #log.info(f'{error_dict}')
                            errs = {category+'_'+key:error_dict[category][key][-1]  for category in error_dict for key in error_dict[category] if category != 'main'}                        
                            errs['main'] = error_dict['main'][-1]
                            xprint('P{}: {} Loop:{} Method:{} Last Errors: \n{}  Best Error: {}\n number of particles = {}, max_density={}'.format(Multiprocessing.get_process_name(),loop_name,iteration,key,errs,state['best_error'],self.rprojection.number_of_particles,np.max(new_density_pair[1])))

                if state['best_iteration']>loop_opt.get('best_density_not_in_first_n_iterations',np.inf):
                    log.info('Selecting density with lowest error metric and continue.')
                    state['density_pair_history']=state['density_pair_history'][1:]+(state['best_density_pair'],)
                    real_pr.support = state['best_mask']
                    state['mask']=state['best_mask']
                state['enforce_initial_support_list']=enforce_initial_support_list
                return state,iteration
            return loop
        loops = [generate_loop_method(name,loop_opt.sub_loops[name],_id) for _id,name in enumerate(loop_opt.sub_loops.order) ]        
        self.loops=loops

        #### assemble main loop ####
        def create_initial_state():
            initial_support = self.projection_objects['real'].initial_support
            real_density_guess = real_density_guess_method()
            #log.info('density guess type = {}'.format(real_density_guess.dtype))
            reciprocal_density_guess=fourier_transform(real_density_guess)
            real_density_guess = inverse_fourier_transform(reciprocal_density_guess)
            #real_density_guess.imag = 0
            #real_density_guess[real_density_guess.real<0]=0
            #real_density_guess[~initial_support]=0
            density_pairs=((reciprocal_density_guess,real_density_guess),)*loop_opt.get('history_length',3)            
                
            error_dict = init_error_dict()

            #first loop
            initial_state = {
                'density_pair_history':density_pairs,
                'error_dict':error_dict,
                'mask':initial_support,
                'best_density_pair':density_pairs[-1],
                'best_error':np.inf,
                'best_iteration':0,
                'best_mask':initial_support}
            return initial_state
        def generate_output(state,iterations,initial_densities,initial_mask):
            best_density_pair = state['best_density_pair']
            best_error = state['best_error']
            best_iteration = state['best_iteration']
            best_mask = state['best_mask']

            output_modifier=routines['output_modifier']
            log.info('last density pair dtypes = {},{}'.format(state['density_pair_history'][-1][0].dtype,state['density_pair_history'][-1][1].dtype))
            out_density_pair = output_modifier.run(*best_density_pair[:2])
            out_last_density_pair = output_modifier.run(*state['density_pair_history'][-1][:2])
            log.info('out last density pair dtypes = {},{}'.format(out_last_density_pair[0].dtype,out_last_density_pair[1].dtype))
            log.info("last density shape = {}".format(out_last_density_pair[1].shape))
            calc_deg2_invariant=routines['calc_deg2_invariant']
            last_deg2_invariant = calc_deg2_invariant.run(out_last_density_pair[1])
            #last_deg2_invariant = calc_deg2_invariant.run(state['density_pair_history'][-1][1])
            
            #log.info("fraction shapes ={}".format([i.shape for i in self.results.get('n_particles_fraction',np.array([]))]))
            error_dict = self.arrayfy_error_dict(state['error_dict'])
            masked_projection_matrices = []
            for mask, matrix in zip(self.rprojection.radial_mask,self.rprojection.projection_matrices):
                tmp = np.array(matrix)
                tmp[~mask]=0
                masked_projection_matrices.append(tmp)
            resultDict={'real_density':out_density_pair[1],
                        'last_real_density':out_last_density_pair[1],
                        'reciprocal_density':out_density_pair[0],
                        'last_reciprocal_density':out_last_density_pair[0],
                        'final_error':best_error,
                        'initial_density':initial_densities[1],
                        'initial_support':initial_mask,
                        'error_dict':error_dict,
                        'support_mask':best_mask,
                        'last_support_mask':state['mask'],
                        'loop_iterations':np.sum(iterations)+1,
                        'fxs_unknowns':self.results['fxs_unknowns'],
                        'n_particles':np.array(self.results.get('n_particles',[])),
                        'n_particles_gradients':np.array(self.results.get('n_particles_gradients',[])),
                        'n_particles_fraction':np.array(self.results.get('n_particles_fraction',[])),
                        "grid_pair":{"real_grid":self.grid_pair.realGrid,
                                     "reciprocal_grid":self.grid_pair.reciprocalGrid},
                        'projection_matrices':masked_projection_matrices,
                        'last_deg2_invariant':last_deg2_invariant}
            return resultDict
        def main_loop(*args,**kwargs):
            initial_state = create_initial_state()
            initial_densities = tuple(d.copy() for d in initial_state['best_density_pair'])
            initial_mask = initial_state['mask'].copy()
            iterations = []
            state = initial_state
            next_state = False
            for loop in loops:
                next_state,iteration = loop(state)
                iterations.append(iteration)
                state = next_state
            out = generate_output(state,iterations,initial_densities,initial_mask)
            return out        
        return main_loop
        
        
    def init_error_dict(self):
        error_names = opt.main_loop.error.methods
        self.results['errors'] = {}
        self.results['errors']['real'] = {name:[] for name in error_names['real']['calculate']}        
        self.results['errors']['reciprocal'] = {name:[] for name in error_names['reciprocal']['calculate']}        
        self.results['errors']['main'] = []
        return self.results['errors']

    def generate_internal_error_limits(self):
        error_methods = opt.main_loop.error.methods
        error_limits = opt.main_loop.error.limits
        error_gain_limits = opt.main_loop.error.gain_limits
        error_names = [*['real_'+key for key in error_methods['real']],
                       *['reciprocal_'+key for key in error_methods['reciprocal']],
                       *['main']]
        
        gain_limits = {name:{phasing_part:lims[_id] for phasing_part,lims in error_gain_limits.items()} for _id,name in enumerate(error_names)}
        limits = {name:lim for name,lim in zip(error_names,error_limits)}
        log.info(limits)
        log.info(gain_limits)
        return limits,gain_limits

    def generate_check_error_routines(self):
        limits,gain_limits = self.generate_internal_error_limits()
        def check_gain_limit(error_dict,loop_method_name):
            return self.check_error_gain_limit(error_dict,gain_limits,loop_method_name)
        def check_limit(error_dict):
            return self.check_error_target_reached(error_dict,limits)
        return check_limit,check_gain_limit
    
    def check_error_gain_limit(self,error_dict,relative_error_limits,loop_method_name):
        error_dict = {**{'real_'+key:err for key,err in error_dict['real'].items()},
                      **{'reciprocal_'+key:err for key,err in error_dict['reciprocal'].items()},
                      **{'main':error_dict['main']}
                      }
        rel_limit_is_reached = False
        for err_name in error_dict:
            limit = relative_error_limits[err_name][loop_method_name]
            if not isinstance(limit , bool):
                errs = error_dict[err_name]
                if len(errs)>=3:
                    rel_err = 1-errs[-2]/errs[-1]
                    #log.info ('\n {} = {} \n'.format(err_name,rel_err))
                    rel_limit_is_reached = (rel_err >= limit) 
                    if rel_limit_is_reached:
                        log.info('Error limit reached: {} change = {}% limit = {}%'.format(err_name,(rel_err)*100,limit*100))
                        break
        return rel_limit_is_reached
    def check_error_target_reached(self,error_dict,error_targets):        
        error_names = opt.main_loop.error.methods
        target_is_reached = False
        error_list_to_short = False
        try:
            error_dict = {**{'real_'+key:err[-1] for key,err in error_dict['real'].items()},
                          **{'reciprocal_'+key:err[-1] for key,err in error_dict['reciprocal'].items()},
                          **{'main':error_dict['main'][-1]}
                          }
        except IndexError as e:
            error_list_to_short = True
        if not error_list_to_short:    
            for err_name in error_dict:
                target = error_targets[err_name]
                if not isinstance(target , bool):
                    err = error_dict[err_name]
                    target_is_reached = (err < target)
                    if target_is_reached:
                        log.info('Target error reached: {}={} limit = {}%'.format(err_name,errs[-1],target))
                        break
        return target_is_reached
                          
    def generate_update_support_mask(self,shrink_wrap_specifier):
        real_pr = self.projection_objects['real']
        def update_real_support_mask(support_mask,*args):
            real_pr.mask = support_mask
        return update_real_support_mask

    def generate_density_guess_method(self,density_guess_specifier,real_grid):
        opt=settings.project.density_guess
        if density_guess_specifier['amplitude_function']=='random':
            def amplitude_function(points):
                np.random.seed(int.from_bytes(os.urandom(4), byteorder='little'))
                return 1+1/opt.random.SNR*np.random.rand(*points.shape[:-1])
        radius = density_guess_specifier['radius']
        if isinstance(radius,bool):
            radius = settings.project.particle_radius
        #log.info('initial density radius {}'.format (radius))
        
        if self.dimensions == 2:                
            integrator = mLib.PolarIntegrator(real_grid[:])
        elif self.dimensions == 3:
            integrator = mLib.SphericalIntegrator(real_grid[:])
            
        total_intensity = self.rprojection.integrated_intensity
        data_type = np.dtype(complex)#settings.project.fourier_transform.data_type

        
        density_type=density_guess_specifier['type']        
        if density_type=='ball':
            if radius<0:
                radius = np.max(real_grid[...,0])
            coord_sys=density_guess_specifier['coord_sys']
            guessed_density_function=SampleShapeFunctions.get_disk_function(radius,amplitude_function=amplitude_function,coordSys=coord_sys)
                
            
            def generate_density():
                #log.info('it is called really')
                density = guessed_density_function(real_grid)
                #scale density to projection data
                total_squared_density = integrator.integrate((density*density.conj()).real)
                #log.info("total squared density = {}".format(total_squared_density))
                #log.info("total intensity = {}".format(total_intensity))                
                density*=np.sqrt(total_intensity/total_squared_density)
                if data_type == 'complex':
                    density = density.astype(complex)
                #log.info('initial total density = {}'.format(integrator.integrate(density**2)))
                return density
        elif density_type=='bump':
            if radius<0:
                radius = np.max(real_grid[...,0])
            slope = opt.bump.slope
            bump_func = mLib.get_test_function(support=[-radius,radius],slope=slope)
            def generate_density():
                A=amplitude_function(real_grid[:])
                #log.info(real_grid[...,0])
                density = A*bump_func(real_grid[...,0])
                #density = bump_func(real_grid[...,0])
                #log.info(density[:,0,0])
                total_squared_density = integrator.integrate((density*density.conj()).real)
                density*=np.sqrt(total_intensity/total_squared_density)
                #ft_density = ft(density.astype(complex))
                #m = ft_density !=0
                #ft_density[m]= ft_density[m]*(np.sqrt(np.abs(I[m]).real)/np.abs(ft_density[m]))
                #density = ift(ft_density)
                if data_type == 'complex':
                    density = density.astype(complex)
                return density
        elif density_type == 'low_resolution_autocorrelation':
            pr = self.projection_objects['reciprocal'].full_projection_matrices
            icht = self.process_factory.get_operator('inverse_harmonic_transform')
            ift = self.process_factory.get_operator('inverse_fourier_transform')
            ft = self.process_factory.get_operator('fourier_transform')
            I = icht(pr)
              
            #thresh = opt.low_resolution_autocorrelation.threshold_to_max
            V_low = self.projection_objects['reciprocal'].low_resolution_intensity_coefficients
            pr = self.projection_objects['reciprocal'].projection_matrices
            V = [np.zeros(p.shape,dtype=complex) for p in pr]
            for vl,v in zip(V_low,V):
                v[:]=vl
            icht = self.process_factory.get_operator('inverse_harmonic_transform')
            ift = self.process_factory.get_operator('inverse_fourier_transform')
            low_res_autocorrelation = ift(icht(pr)).real
            low_res_autocorrelation[low_res_autocorrelation<0]=0
            _max = low_res_autocorrelation.max()
            low_res_density_guess = low_res_autocorrelation
            mean_density = low_res_density_guess.mean()
            #db.save("/gpfs/exfel/theory_group/user/berberic/MTIP/test/inital_density/d1_low_res.vts",[np.abs(low_res_density_guess)],grid_type='spherical',grid = real_grid[:])
            radius = settings.project.particle_radius
            bump_func = mLib.get_test_function(support=[-radius,radius],slope=0.1)
            def generate_density():
                A=amplitude_function(real_grid[:]).astype(complex)
                density = low_res_density_guess*A
                density[density<0]=0
                density*=bump_func(real_grid[...,0])
                total_squared_density = integrator.integrate((density*density.conj()).real)
                density*=np.sqrt(total_intensity/total_squared_density)
                return density
        else:
            e=AssertionError('density type "{}" is not known.'.format(density_type))
            log.error(e)
            raise e
        return generate_density

    def generate_update_shrink_wrap(self,sw_object):        
        default_sigma = sw_object.default_sigma
        opt = settings.project
        loop_opt = opt.main_loop
        sw_opt = opt.projections.real.shrink_wrap
        
        nloops = len(loop_opt.sub_loops.order)
        sw_sigma_ramps = []
        sw_thresh_ramps = []
        
        for lid in range(nloops): 
            n_sw_sigmas = len(sw_opt.sigmas)
            n_sw_thresholds = len(sw_opt.thresholds)
            if n_sw_sigmas-1 < lid:
                sw_sigma = False
                log.info(f'No SW sigma value in settings for loop with id = {lid} in {loop_opt.sub_loops.order} specified. Settings sigma to: {sw_sigma} ')
            else:
                sw_sigma = sw_opt.sigmas[lid]
            if not isinstance(sw_sigma,(list,tuple)):
                sw_sigma = [sw_sigma]
            sw_sigma_ramp = LinearRamp(*sw_sigma,default_start = sw_object.default_sigma,default_stop = sw_object.default_sigma)
            sw_sigma_ramps.append(sw_sigma_ramp)
            #log.info(f'{loop_name} : sigma = {sw_sigma} from simgas = {sw_opt.sigmas}')
            
            n_sw_thresholds = len(sw_opt.thresholds)            
            if n_sw_thresholds-1 < lid:
                sw_threshold = 0.1
                log.info(f'No SW threshold value in settings for loop with id = {lid} in {loop_opt.sub_loops.order} specified. Settings threshold to: {sw_threshold}')
            else:
                sw_threshold = sw_opt.thresholds[lid]

            if not isinstance(sw_threshold,(list,tuple)):
                sw_threshold = [sw_threshold]
            sw_thresh_ramp = LinearRamp(*sw_threshold)
            sw_thresh_ramps.append(sw_thresh_ramp)
        def update_shrink_wrap(iteration,loop_number):
            sw_sigma_ramp = sw_sigma_ramps[loop_number]
            if not sw_sigma_ramp.undefined:
                sw_object.gaussian_sigma = sw_sigma_ramp(iteration)
            sw_threshold_ramp = sw_thresh_ramps[loop_number]
            if not sw_threshold_ramp.undefined:
                sw_object.threshold = sw_threshold_ramp(iteration)
    
            #log.info(f'In {iteration}-th SW application of loop {loop_number} : sigma = {sw_object.gaussian_sigma}, threshold = {sw_object.threshold}')
            

        return update_shrink_wrap
    def arrayfy_error_dict(self,error_dict):
        for category in error_dict:
            if category == 'main':
                error_dict[category] = np.array(error_dict[category])
            else:
                for err_key,errs in error_dict[category].items():
                    error_dict[category][err_key] = np.array(error_dict[category][err_key])
        return error_dict


    def generate_phasing_loop(self):
        #self.load_mtip_data()
        log.info('Setting up operators')
        operators,grid_pair=self.assemble_operators()
        self.process_factory.addOperators(operators)
        self.grid_pair = grid_pair
        log.info('Setting up phasing routines (HIO,ER,etc.)')
        self.routines = self.assemble_MTIP_routines()
        log.info('Assemble phasing loop (HIO,ER,etc.)')
        self.phasing_loop = self.assemble_phasing_loop()



import numpy as np
from xframe.library import mathLibrary as mLib
from xframe import log
import scipy as sp
log=log.setup_custom_logger('root','INFO')

#A = np.random.rand(3*20*10).reshape(3,20,10)
size = 100
A = np.diag(np.exp(-np.arange(0,size)*np.log(1e+5)/size))
A[0,-1]=100
x_exact = np.arange(0.9,1,0.1/size)[::-1]
b_free = A.dot(x_exact)
b = b_free + np.random.normal(scale = 0.001,size = size)
b = np.stack((b))



'''
    Finds x that minimizes $|| A x -B ||^2 - ||\lambda Id x ||^2 $
    '''
lambd = 0.617
n_orders = A.shape[-1]
lambd = np.atleast_1d(lambd)
if A.ndim < 3:
    A = A[None,...]
    b = b[None,...]
#n_qs = A.shape[0]
#log.info('n_ord = {} n_qs = {}'.format(n_orders,n_qs))
#eyes = np.asarray(lambd)[:,None,None]*np.eye(n_orders)[None,...]
#log.info('A shape = {} eyes_shape = {}'.format(A.shape,eyes.shape))
#new_A = np.concatenate((A,eyes),axis = 1)
#new_b = np.concatenate((b,np.zeros((n_qs,n_orders))),axis = 1)

AA = np.copy(A)
bb = np.copy(b)
#x = np.sum(vh*d_UT_y[:,:,None],axis = -2)
#preprocess data:
A_offset = np.average(A[0], axis=0,weights=None)
A_offset = A_offset.astype(A[0].dtype, copy=False)
AA[0] -= A_offset

A_scale = np.ones(A[0].shape[1], dtype=A[0].dtype)
b_offset = np.average(b[0], axis=0,weights=None)
bb[0] -= b_offset

try:
    #lambd = mLib.approximate_tikhonov_parameters(AA,bb)
    lambd = mLib.approximate_tikhonov_parameters(A,b)
    #lambd = np.array([0.0])
except Exception:
    print('fuu')
    lambd = np.array([0.0])
U, s, Vt = sp.linalg.svd(A[0], full_matrices=False)
idx = s > 1e-15  # same default value as scipy.linalg.pinv
s_nnz = s[idx][:,None]
UTy = np.dot(U.T, b[0])
d = np.zeros((s.size, lambd.size), dtype=A[0].dtype)
d[idx] = s_nnz / (s_nnz**2 + lambd)
d_UT_y = d * UTy[:,None]
xx = np.dot(Vt.T, d_UT_y).T
x = mLib.tikhonov_solver_svd(A,b,lambd)

intercept = b_offset - np.dot(A_offset, x.T)

bx = A[0].dot(x[0])#+b_offset#A[0].dot(x[0])+intercept
#bx = A[0].dot(x[0])+intercept


from sklearn import linear_model
reg = linear_model.RidgeCV(alphas=np.logspace(-6, 6, 13),gcv_mode='svd',fit_intercept=False,normalize = False)
reg2 = linear_model.Ridge(alpha=0,solver='svd',fit_intercept=False,normalize = False)
reg.fit(A[0],b[0])
reg2.fit(A[0],b[0])
x1 = reg.coef_
x2 = reg2.coef_
bbx = A[0].dot(reg2.coef_)+reg2.intercept_

err = np.mean(np.linalg.norm(bx-b_free)/np.linalg.norm(b_free))
err2 = np.linalg.norm(np.abs(bbx-b_free)/np.linalg.norm(b_free))
xerr = np.mean(np.linalg.norm(x-x_exact)/np.linalg.norm(x_exact))
xerr2 = np.linalg.norm(np.abs(x1-x_exact)/np.linalg.norm(x_exact))
h=np.linalg.lstsq(A[0],b[0])[0]
xerr3 = np.linalg.norm(np.abs(h-x_exact)/np.linalg.norm(x_exact))
print('own',err)
print(err2)
print(reg.alpha_)
print(lambd)
print(xerr)
print(xerr2)
print(xerr3)
#x = np.array(tuple(np.linalg.lstsq(A_part,b_part,rcond = None)[0] for A_part,b_part in zip(new_A,new_b)))

    



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_associated_legendre_matrices
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi,plot1D
from xframe import Multiprocessing
from scipy.interpolate import interp1d
from scipy.linalg import solve_triangular
from scipy.stats import unitary_group

log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
from xframe.library.mathLibrary import gsl
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
#analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_10p_spher')[0] 
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
controller.chooseJob()
ccd_1 = db.load('ccd')


def init_transforms():
    cc_data = ccd_1['cross_correlation'][...,:800]
    maxQ = np.max(ccd_1['radial_points'])*2
    max_order = 99
    n_radial_points = 128#300
    opt=settings.analysis
    ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':False,'n_theta':False}
    #ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
    # harmonic transforms and grid
    cht=HarmonicTransform('complex',ht_opt)
    weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points,expansion_limit = 2000)
    grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':maxQ,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
    grid_pair = get_grid(grid_opt)
    maxR = grid_pair.real[:,0,0].max()
    ft,ift = generate_zernike_spherical_ft(maxR,weight_dict,cht,'complex',True,use_gpu=True)
    return ft,ift,cht,grid_pair

#locals().update(init_transforms())

#projd = db.load('reciprocal_proj_data',path_modifiers={'name':'3d_model_0','max_order':str(99)})

#p_matrices = list(projd['data_projection_matrices'])
#for order in range(projd['max_order']+1):
#    shape = p_matrices[order].shape
#    if 2*order+1 > shape[1]:
#        tmp = np.zeros((shape[0],2*order+1))
#        tmp[:,:shape[1]] = p_matrices[order]
#        p_matrices[order] = tmp 
scan_range = [1,20.1,0.5]
names = [
    '3d_pentagon_100_mO99_inverse_binned_psd20_arc_wrong_n_particles',
    '3d_pentagon_10p_spher_mO99_inverse_binned_psd20_arc_wrong_n_particles',
    '3d_pentagon_10p_spher_mO99_inverse_binned_psd20_custom_wrong_n_particles',
    '3d_model_0_mO99_cn_inverse_not_binned_psd20_custom',
    '3d_pentagon_10p_spher_mO99_inverse_binned_psd20_custom',
    '3d_pentagon_10p_noise_mO99_inverse_arc_no-psd',
    '3d_pentagon_Unknown_p_noise_mO99_inverse_arc_no-psd',
    '3d_model_3_nosym_mO99_psd_eigh'
]
cc_names = [
    'pentagon_100_particles',
    '3d_pentagon_10particles_spherSim',
    '3d_pentagon_10particles_spherSim',
    '3d_model_0',
    '3d_pentagon_10particles_spherSim',
    '3d_pentagon_10particles_noise',
    '3d_pentagon_?particles_noise',
    '3d_model_3'
]
shift = 0
n_points = 40
scales = np.logspace(np.log10(1/(10-shift+1)),np.log10(10+shift-1),n_points) #np.arange(*scan_range)
scales = np.linspace(1,10,n_points)
scales_step = scales[1]-scales[0]
radius = 250.0
select  = -3
name = names[select]
cc_name = cc_names[select]

def scan_n_particles(scales,dataset_name,scan_range,radius,shift,**kwargs):
    mtip = controller.job.mtip
    mtip.generate_phasing_loop()
    grid_pair = mtip.grid_pair
    r = mtip.grid_pair.realGrid[:,0,0,0]
    q = mtip.grid_pair.reciprocalGrid[:,0,0,0]
    #ft,ift,cht,grid_pair = init_transforms()
    #projd = db.load('/gpfs/exfel/theory_group/user/berberic/MTIP/reciprocal_proj_data/'+dataset_name+'.h5')
    p_matrices = mtip.rprojection.projection_matrices #list(projd['data_projection_matrices'][str(i)] for i in range(projd['max_order']+1))
    #p_matrices[0]= p_matrices[0].reshape(-1,1)
    max_order = len(p_matrices)-1
    ft =mtip.process_factory.operatorDict['fourier_transform']
    cht = mtip.process_factory.operatorDict['harmonic_transform']
    icht = mtip.process_factory.operatorDict['inverse_harmonic_transform']

    spherical_formfactor = pLib.spherical_formfactor(q,radius=radius)
    neg = []
    n_pixel_in_radius = np.sum(r<=radius)
    base_guess = np.ones(grid_pair.realGrid.shape,dtype = float)
    base_guess[r>radius]=0
    I_base_guess = np.zeros_like(base_guess)
    I_base_guess[:] = spherical_formfactor[:,None,None]
    np.random.seed(int.from_bytes(os.urandom(4), byteorder='little'))
    for repeats in range(1):        
        neg_fractions = []
#        log.info('repeat count = {}'.format(repeats))
        for scale in scales:
            #log.info('creating density')
            #density_guess = np.array(base_guess)
            #density_guess[:n_pixel_in_radius]+= 0.1*np.random.rand(n_pixel_in_radius,*base_guess.shape[1:])
            #log.info('finished density of shape = {}'.format(density_guess.shape))
            #I_guess = np.abs(ft(density_guess))**2
            I_guess = np.array(I_base_guess)
            I_lm_guess = cht(I_guess)
            I_lm_guess[1:] = [ np.zeros_like(i) for i in I_lm_guess[1:]] 
            unknowns = mtip.process_factory.operatorDict['approximate_unknowns'](I_lm_guess)
            #log.info(unknowns[-1].shape)
            #I_lm = [ np.sum(p_matrices[order][:,:,None] * unknowns[order][None,:,:],axis = 1) for order in range(max_order+1)]
            I_lm = [ np.zeros_like(i) for i in I_lm_guess]
            for i in range(len(I_lm)):
                p_matrix = p_matrices[i]
                I_l = I_lm[i]
                I_l[:,:p_matrix.shape[-1]]=p_matrix
                
            I_lm[1:] = [lm/shift for lm in I_lm[1:]]
            I_lm[0]/=scale
            I = icht(I_lm)
            neg_fractions.append(np.sum(I.real<0)/np.prod(I.shape))
        neg.append(np.array(neg_fractions))
    return np.mean(neg,axis = 0)

def init_2():
    pd = db.load('/gpfs/exfel/theory_group/user/berberic/MTIP/reciprocal_proj_data/'+name+'.h5')
    ccd = db.load('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/'+cc_name+'.h5')
    max_order = pd['max_order']
    p_matrices = [pd['data_projection_matrices'][str(o)] for o in range(max_order +1)]
    p_matrices[0] = p_matrices[0][:,None] 
    qs = pd['data_radial_points']
    nq = len(qs)
    cns = i_tools.deg2_invariant_to_cn_3d(pd['deg_2_invariant'],qs,pd['xray_wavelength'])
    log.info('finished')
    bl = np.moveaxis(pd['deg_2_invariant'],0,-1)
    cc = ccd['intra']['ccf_2p_q1q2']
    hcc = ccd['intra']['fc_2p_q1q2']
    
    aint = pd['average_intensity']
    
    return locals()
locals().update(init_2())

thetas = pLib.ewald_sphere_theta(pd['xray_wavelength'], pd['data_radial_points'])
qq_matrix = np.squeeze(i_tools.ccd_associated_legendre_matrices(thetas,99,0))

bl_new = bl.copy()
scales = np.linspace(1,100,200)
bl0 = aint[:,None]*aint[None,:]*4*np.pi
cc0 =(qq_matrix[...,0]*bl[...,0])[...,None]/scales[None,None,:]**2+np.sum(qq_matrix[...,1:]*bl_new[...,1:],axis = -1)[:,:,None]
#cc0 =(qq_matrix[...,0]*bl0)[...,None]/np.sqrt(scales[None,None,:])+np.sum(qq_matrix[...,1:]*bl_new[...,1:],axis = -1)[:,:,None]
#cc0 = np.moveaxis(cc0,-1,0)[:,20:,20:]
qs=pd['data_radial_points'][20:]

neg_sum = []
temp = np.zeros_like(cns[20:,20:])
temp[...,:100] = cns[20:,20:]
tot = len(qs)**2 * cc.shape[-1]
for s in scales:
    #temp[...,0]= (qq_matrix[...,0]*pd['deg_2_invariant'][0]/s**2+np.sum(qq_matrix[...,1:]* np.moveaxis(pd['deg_2_invariant'],0,-1)[...,1:],axis = -1))[20:,20:]
    #temp[...,0]= (qq_matrix[...,0]*pd['deg_2_invariant'][0]/s**2+2*np.sum(qq_matrix[...,1:]* np.moveaxis(pd['deg_2_invariant'],0,-1)[...,1:],axis = -1))[20:,20:]
    temp[...,0]= (qq_matrix[...,0]*bl0/s**2+8*np.sum(qq_matrix[...,1:]* np.moveaxis(pd['deg_2_invariant'],0,-1)[...,1:],axis = -1))[20:,20:]
    cc = np.fft.irfft(temp,1600) #mLib.circularHarmonicTransform_real_inverse(temp,200)
    temp[...,1:]=cns[20:,20:,1:]*8
    neg_sum.append(np.sum(cc<0)/tot)

#cc0 = np.moveaxis(cc0,-1,0)[:,20:,20:]
#qs= qs[20:]
neg_mask = cc0<0
#neg_sum = np.array([ np.sum(-1*cc0[s,neg_mask[s]]*((qs**2)[:,None]*(qs**2)[None,:])[neg_mask[s]]) for s in range(len(scales))])
#neg_sum = np.sum(neg_mask*((qs**2)[:,None]*(qs**2)[None,:])[None,...],axis = (1,2))
#neg_sum = np.sum(neg_mask[20:,20:],axis = (0,1))
neg_sum = np.array(neg_sum)
#neg_sum = np.sum(neg_mask,axis=(0,1))
grad = np.gradient(neg_sum,scales[1]-scales[0])

pl_name = 'bl_c0_test'
def print_neg():
    expected=100
    #expected=shift
    grid = scales
    layout = {'title':'{} \n Assumed particle radius ={}'.format(name.replace('_','\_'),radius),'x_label':'scaling factor', 'y_label':'\% of negative pixels in I'}#,))*len(orders)
    fig = plot1D.get_fig(neg_sum,grid = grid,x_scale='lin',layout = layout)
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/n_particles/{}_new3.matplotlib'.format(pl_name),fig)
    layout = {'title':'{} \n Assumed particle radius ={}'.format(name.replace('_','\_'),radius),'x_label':'scaling factor', 'y_label':' diff '}#,))*len(orders)
    #diff = (neg[:-2]-neg[2:])/(scales[:-2]-scales[2:])
    fig = plot1D.get_fig(grad,grid = scales[:],x_scale='lin',layout = layout)
    ax = fig.get_axes()[0]
    #ax.vlines(expected,grad.min(),grad.max())
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/n_particles/{}_diff_new4.matplotlib'.format(pl_name),fig)
print_neg()



import numpy as np
import os
import glob
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi,heat2D,heatPolar2D,agipd_heat
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward,cartesian_to_spherical,spherical_to_cartesian
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward,masked_mean
from xframe.presenters.matplolibPresenter import plot1D,hist2D,heat2D



import logging
#from xframe.plugins.fxs3046_online.analysisLibrary.regrid2 import SimpleRegridder2D,AgipdRegridderSimple

from scipy.spatial.kdtree import KDTree
from scipy.ndimage import gaussian_filter

log=logging.getLogger('root')

init  = True
if init:
    from xframe.startup_routines import dependency_injection
    dependency_injection()
    from xframe.startup_routines import load_recipes
    from xframe.control.Control import Controller
    analysis_sketch,exp_sketch = load_recipes(ana_name='fxs3046_online.generate_plots_2', ana_settings='saxs_65_-bg',exp_name='fxs3046_online',exp_settings = 'default_proc')
    from xframe import database
    #controller = Controller(analysis_sketch,experiment_sketch = exp_sketch)
    #controller.chooseJobSingleProcess()
    #from xframe import database
    #from xframe import Multiprocessing
    #experiment = Multiprocessing.comm_module.get_experiment()
    #aw = controller.controlWorker.analysis_worker
    #aw.geometry = aw.load_geometry()
    #aw.agipd_regridder = False
    #r = controller.controlWorker.analysis_worker.set_and_get_agipd_regridder()

    

db = database.experiment
db.create_vds(1,False)
